#! /usr/bin/perl -w
# betascan.pl
# main driver for Betascan
# Allen Bryan 10/27/08


##################################################################
### 0. Configuration notes and summary of changelog            ###
##################################################################

### 0.1 Configuration notes ###

# BETASCAN uses the Perl matlab connector, Math::Matlab 0.07
# to remotely invoke MATLAB.
# The default location is '/Library/Perl/Math-Matlab-0.07/lib'
# change this under 1.1 Perl dependencies if your matlab is failing
# and no images are being generated.

# Please note that we have had problems with the 2008 release
# of MATLAB not having all necessary libraries for image generation
# in the standard install. 2007 should still work. Bug reports are 
# welcome and requested.

# The location of MATLAB itself is defaulted in $matlabLoc, 
# but can be input at command line with switch -G.
# Feel free to alter $matlabLoc if switch use becomes painful.

# BSprobs.txt and BSweights.txt must be in the same directory 
# with betascan.pl.

# Please email betascan@csail.mit.edu
# for questions, comments, concerns, bug reports, and all other
# correspondence on use of BETASCAN. Good night, and good luck.

### 0.2 Changelog summary ###

# version 6.1: 
# uses amphipathic tables
# allows bonuses; allows sequences to be run against each other
# this version uses matlab to generate singleton and pairwise plots
# see POD for input switches
# does maxima finding and uses matlab to generate plots
# generates html file as well
# matlab files autogenerated
# all files use corrector
# optimized for batch use: program falls through rather than fails on no results

# uses subsidiary programs and routines:
# * calcProbs() was code4.cpp (Matt Menke)
# * 3dranger() (Allen Bryan)
# * singleReformatSwap() (Allen Bryan)
# * readWeightFiles() was from okolona.pl (Andrew McDonnell)
# * bsplotsingle.m (Allen Bryan) (now autogenerated!)
# * bsplotpair.m (Allen Bryan) (now autogenerated!)
# * bsbetasingle.m (Allen Bryan) (now autogenerated!)
# * bsbetapair.m (Allen Bryan) (now autogenerated!)

##################################################################
### 1. Initializations                                         ###
##################################################################

### 1.1 Perl dependencies ###

use lib '/Library/Perl/Math-Matlab-0.07/lib';
use Math::Matlab::Local;
use Cwd;
use Getopt::Std;
use strict;

#handles to force printing as quickly as feasible
use IO::Handle;
STDIN->autoflush(1);
$| = 1;

### 1.1.1 Get the time ###

my $runTime = time;
my $idTime = ($runTime % 1000000);
my $timeString = localtime($runTime);

print STDERR "1.\tWelcome to BETASCAN, the time is $timeString. \n";

### 1.1.2 Error file ###

my $errorFile = "Betascan-errors-$idTime.txt";
open(MAINERR, ">$errorFile") or die "Betascan dies: cannot open sequence error file $errorFile";
print MAINERR "BETASCAN run for $timeString \n";

### 1.2 Command line argument read-in ### 

my @data_files = @ARGV;
my $opts = {};
my $optString = "";
getopts('n:b:l:m:M:N:i:c:w:p:j:a:z:C:G:12dsfgohFDtyALTZe', $opts);

# shut down if mandatory switches are missing
&usage if !defined $opts->{'n'};

### 1.3 Process switches ###

### Procedure for adding a new switch:   ###
### Code it here to get it in the system ###
### Code it to frob the frobbable thing  ###
### If required, report it in the html   ###
### Add it to the POD at the end of file ###


### Switches which control global program behavior ###

# 1.3.17 -j: define a job ID from outside
if (defined $opts->{'j'}) {
	$idTime = $opts->{'j'};
}

# 1.3.20 -D: debug mode
# hardcoded

# 1.3.last? -C: cache address for server use
my $address = 'DEADBEEF';
if (defined $opts->{'C'}) {
	$address = $opts->{'C'};
}

# 1.3.reallylast: -G: matlab location
my $matlabLoc = "/Applications/MATLAB701/bin/matlab";
if (defined $opts->{'G'}) {
	$matlabLoc = $opts->{'G'};
}

### Switches receiving global program data ###

# 1.3.1 -n: name
my $inputFile = $opts->{n};
print MAINERR "Input file: $inputFile \n";
print "INPUT FILE = $inputFile \n";

# 1.3.2 -b: beta strand length
my $maxBetaLength;
if (defined $opts->{'b'}) {
	$maxBetaLength = $opts->{'b'};
	int($maxBetaLength) or die "Betascan dies: $maxBetaLength can't be a length!\n";
	($maxBetaLength > 1) or die "Betascan dies: a beta strand can't be $maxBetaLength long!\n";
	($maxBetaLength < 20) or die ("Betascan dies: I'm sorry, Dave, but I'm afraid I can't do that." . 
									" There are no strands of length $maxBetaLength in the PDB\n");
}
else { $maxBetaLength = 13; }
print "BETA LENGTH = | $maxBetaLength | \n";
$optString .= "-b $maxBetaLength ";

# 1.3.3 -l: sequence length
my $maxSeqLength = -1;
if ( defined $opts->{l} ) {
	$maxSeqLength = $opts->{l};
	# print $maxSeqLength;
	int ($maxSeqLength) or die  "Betascan dies: $maxSeqLength can't be a sequence length!\n";
	# print $maxSeqLength;
	($maxSeqLength > 10) or die "Betascan dies: $maxSeqLength is way too short!\n";
	($maxSeqLength < 300) or print STDERR "Warning: Sequence lengths over 300 may take a while to process.\n";
}
$optString .= "-l $maxSeqLength ";

# 1.3.4 -m: minimum loop length
my $minLoop;
if (defined $opts->{'m'}) {
	$minLoop = $opts->{m};
	int ($minLoop+1) or die "Betascan dies: minimum loop length must be zero or positive integer\n";
	$optString .= "-m $minLoop ";
}
else {
	$minLoop = 0;
}

# 1.3.5 -M: maximum loop length
my $maxLoop;
if (defined $opts->{'M'}) {
	$maxLoop = $opts->{M};
	if ($maxLoop == 0) {
		$maxLoop = 1000000;
	}
	else {
		int ($maxLoop) or die "Betascan dies: maximum loop length must be an integer. Set to zero for no maximum.\n";
		($maxLoop >= $minLoop) or die "Betascan dies: maximum loop must be at least as long as minimum loop! \n";
	}
	$optString .= "-M $maxLoop ";
}
else {
	$maxLoop = 35;
}	

### Switches globally controlling singleton and pairwise calculation ###

# 1.3.23 -A: antiparallel calculations included
# hardcoded

# 1.3.15 -a: two-sequence probabilities
# need a switch for option to process two sequences against one another
my $alignTwoName = 0;
if (defined $opts->{'a'}) {
	$alignTwoName = $opts->{'a'};
}
# -a still needs implementing

### Switches affecting direct singleton and pairwise probabilities ###

# 1.3.13 -p: probability file name
# be able to frob the probabilities to something different
my $probFileName = "BSprobs.txt";
if (defined $opts->{'p'}) {
	$probFileName = $opts->{'p'};
	if ($probFileName ne "BSprobs.tst") {
		$optString .= "-p $probFileName ";
	}
}

# 1.3.22 -y: tyrosine and other stacking bonuses
if (defined $opts->{'y'}) {
	$optString .= "-y ";
}
# implemented by hardcoding; consider using file

### Switches affecting sequence correction of probabilities ###

# 1.3.11 -c: specify model
# turn on use of log-odds scores rather than direct probabilities
# options to implement: null, allproteins, allprions, yeastproteins, yeastbetas, yeastprions
my $correctFlag = 0;
my $model = "allproteins";
if (defined $opts->{'c'}) {
	$correctFlag = 1;
	$model = $opts->{'c'};
	print STDERR "1.3.11\tUsing amino acid model $model\n";
	$optString .= "-c $model ";
}

# 1.3.12 -w: weight file name
# be able to frob the name of the weight file
my $weightFileName = "BSweights.txt";
if (defined $opts->{'w'}) {
	$weightFileName = $opts->{'w'};
	if ($weightFileName ne "BSweights.txt") {
		$optString .= "-w $weightFileName ";
	}
}

### Switches affecting length normalization ###

# 1.3.6 -N: normalization parameter
# this is the parameter to the Poisson distribution that describes
# the length distribution of parallel beta strands
my $normParam;
if (defined $opts->{'N'}) {
	$normParam = $opts->{'N'};
	$optString .= "-N $normParam ";
}
else {
	$normParam = 3.15;
}

# 1.3.24 -L: turn off length normalization
# this might (or might not!) be a good idea for the stitcher
if (defined $opts->{'L'}) {
	$optString .= "-L ";
}
# hardcoded

### Switches affecting other normalizations ###

# 1.3.10 -o: turn on orientation normalization, something that has been questioned
my $orientFlag = 0;
if (defined $opts->{'o'}) {
	$orientFlag = 1;
	$optString .= "-o ";
}

### Switches affecting image generation ###

# 1.3.7 -i: image format control 
# make images controllable by an option switch too
my $imageOpt;
if (defined $opts->{'i'}) {
	$imageOpt = $opts->{'i'};
	if ($imageOpt ne "all" && $imageOpt ne "jpg" && $imageOpt ne "tiff" && $imageOpt ne "pdf" && $imageOpt ne "eps") {
		print STDERR "Betascan dies: invalid image option. choose: all, jpg, pdf, tiff, eps.";
	}
}
else {
	$imageOpt = "pdf";
}

# -i still has matlab problems...

# 1.3.18 -g: turn off all graphics generation
# hardcoded

# 1.3.25 -s: turn off all graphics generation and calculations
# hardcoded

# 1.3.8 -1, -2: singleton and pairwise switches
# turn off singleton or pairwise image generation
my $genFlag = 0;
if (defined $opts->{'1'}) {
	$genFlag = 1;
}
if (defined $opts->{'2'}) {
	$genFlag = 2;
}

# 1.3.whatever -z: shifted counting point
# allows user to shift the starting point of the residue numbering

my $countShift = 0;
if (defined $opts->{'z'}) {
	$countShift = $opts->{'z'};
}

# 1.3.16 -F: calculate fast, only the requireds
# hardcoded

# 1.3.whocares -T: large type for journal images
my $largePrint = 0;
if (defined $opts->{'T'}) {
	$largePrint = 1;
}

### Switches affecting processing ###

# 1.3.19 -f: turn on maxima generation
# hardcoded

# 1.3.pleasebelast -e: eliminate poor maxima with filter system
my $filter = 0;
if (defined $opts->{'e'}) {
	$filter = 1;
}

# 1.3.14 -h: html output
# frob off a html file that links the pages you want to see
my $htmlFlag = 0;
my $htmlFileName;
unless (defined $opts->{'h'}) {
	$htmlFlag = 1;
	if (defined $opts->{'C'}) {
		$htmlFileName = $address . "betascan-". $idTime . ".html";
	}
	else {
		$htmlFileName = "betascan-" . $idTime . ".html";
	}
}

### Switches affecting file organization ###

# 1.3.9 -d: turn off/on directory sorting
my $dataFlag = 1;
if (defined $opts->{'d'}) {
	$dataFlag = 0;
}

# 1.3.random -Z: tar and zip all outputs (hardcoded)

### Time debug switch: must go last ###

# 1.3.21 -t: time debug mode for code optimization

my %timeRef;
my $timeFlag = 0;
if (defined $opts->{'t'}) {
	$timeFlag = 1;
	$timeRef{"01.3 switch"} = time - $runTime;
}

### 1.4 Generate matlab processing files ###

my $matlabFileName = generateMatlab($idTime, $imageOpt);

### 1.5 All switches done ###

print MAINERR "Switch set: $optString \n";
print MAINERR "All switches processed.\n";

##################################################################
### 2. Probability Calculations                                ###
##################################################################

print STDERR "2.\tCalculating beta strand probabilities.\n";

### 2.1 Read in sequence names ###
# Note that currently, sequence names are not allowed to have spaces!
# This may be deprecated soon because calcProbs reads the file names too 

my $seqHashRef;
my $seqNameHashRef;
print MAINERR "\tCalling readSequence: ";
($seqHashRef, $seqNameHashRef) = readSequence($inputFile);
if ($timeFlag) {
	$timeRef{"02.1 readSequence"} = time - $runTime;
}
my @proteinNames = keys (%$seqHashRef);

# How about if I read the weights file only once??
my $weights = readWeightFile($weightFileName);

### 2.2 Raw probability calculations ###
# now using shiny new perl translation!
print MAINERR "\tCalling calcProbs: ";
calcProbs($seqHashRef, $maxBetaLength, $minLoop, $maxLoop, $normParam, $model);
if ($timeFlag) {
	$timeRef{"02.2 calcProbs"} = time - $runTime;
}

##################################################################
### 3. Graphics Processing                                     ###
##################################################################

### 3.1 Matlab and filename initializations ###

my $cwd = getcwd();
my $matlab;

unless (defined $opts->{'g'}) {
	print STDERR "3.1\tInvoking the Matlab engine.\n";
	print MAINERR "Invoking the Matlab engine.\n";
	
	$matlab = Math::Matlab::Local->new({
		cmd => "$matlabLoc -nodisplay -nojvm -logfile matlab-$idTime.log",
		root_mwd => $cwd});
		
	if ($timeFlag) {
		$timeRef{"03.1 Matlab start"} = time - $runTime;
	}
}
	
my @basedranger = ("prion combined scores", "prion forward scores");

if (defined $opts->{'A'}) {
	push(@basedranger, "antiparallel prion combined scores");
	push(@basedranger, "antiparallel prion forward scores");
}

my @basereformat = ("prion single strand");
my @orient = ("in-out", "out-in");
my $kk;

my @length = ();
for (my $jj = 2; $jj <= $maxBetaLength; $jj++) {
	push(@length, $jj);
}
my $suffix = "residue.txt";
my $swapSuffix = "complete swap.txt";
my $mybase = '';
my $currentInName = '';

### 3.2 Main loop begins ###

print STDERR "3.2\tGenerating tables and graphs:\n";
print MAINERR "Begin main loop\n";
if ($timeFlag) {
	$timeRef{"03.2 Loopstart"} = time - $runTime;
}
foreach my $mySpecies (@proteinNames) {
	
	print STDERR "\t\tFor $mySpecies:\n";
	print MAINERR "For $mySpecies:\n";
	my $mySequence = $seqHashRef->{$mySpecies};
	my $lengthToUse = $maxSeqLength;
	if ($lengthToUse < 0) {
			$lengthToUse = length($seqHashRef->{$mySpecies});
	}
	print MAINERR "\tSequence = $mySequence \n";
	print MAINERR "\tusing length $lengthToUse \n";

	if (!defined $opts->{'s'}) {
	
### 3.2.1: Pairwise log-odds scores and graphics ###

		# if not singleton-only, do
		unless ($genFlag == 1) { 	
			print MAINERR "Doing pairwise calcuations\n";	
			print STDERR "3.2.1\t\tDoing pairwise calculations\n";	
			foreach $mybase (@basedranger) {
			print STDERR "\t\t\tDoing $mybase - ";
			print MAINERR "\tDoing $mybase\n";
				foreach my $myOrient (@orient) {
					print STDERR "$myOrient";
					print MAINERR "\t\t$myOrient\n";
					foreach my $myLength (@length) {
								print STDERR ".";
								my $twoDigit = sprintf("%02d", $myLength);
								$currentInName = "$mySpecies $mybase $myOrient $twoDigit $suffix";
								my $currentOutName = "$mySpecies 3dranged $mybase $myOrient $twoDigit $suffix";
								if (defined $opts->{'D'}) {
									print MAINERR "\t\t\tCalling threeDRanger with arguments: \n";
									print MAINERR "\t\t\t\t$currentInName,\n ";
									print MAINERR "\t\t\t\t$currentOutName,\n ";
									print MAINERR "\t\t\t\t$myLength, $lengthToUse, \n";
									print MAINERR "\t\t\t\t$mySequence\n";
								}
								my $combinedFlag = 0;
								if ($mybase eq "prion combined scores") {
									if (defined $opts->{'D'}) {
										print MAINERR "\t\t\t\tCOMBINED! \n";
									}
									$combinedFlag = 1;
								}
								
								threeDRanger($currentInName, $currentOutName, $myLength, 
												$lengthToUse, $mySequence, $combinedFlag);
												
								unless (defined($opts->{'g'}) || !defined($opts->{'F'})) {
									my $currentMatlabExecutable = "bsplotpair400$idTime('$currentOutName', '$optString', $countShift)";
									print MAINERR "\t\t\tCalling Matlab: $currentMatlabExecutable\n";
									my $matlabOutput = $matlab->execute($currentMatlabExecutable);							
									print MAINERR "\t\t\t\tResult: $matlabOutput\n";
								} # end unless
                        } # end foreach $myLength 
                } # end foreach $myOrient
			print STDERR "\n";
			}	# end foreach $mybase
			if ($timeFlag) {
				$timeRef{"03.2.1 pairwise $mySpecies"} = time - $runTime;
			}
		} # end if genFlag != 1
	
### 3.2.2: Singleton log-odds scores and graphics ### 

		# if not pairwise-only, do
		unless ($genFlag == 2) {
			print MAINERR "Doing singleton calculations\n";
			foreach $mybase (@basereformat) {
				if ( !defined($opts->{'F'}) && $mybase ne "prion single strand") {
					next;
				}
				print STDERR "3.2.2\t\tReformatting $mybase...";
				$currentInName = "$mySpecies $mybase";
				if (defined $opts->{'D'}) {
					print MAINERR "\tCalling singleReformatSwap with arguments: \n";
					print MAINERR "\t\t$currentInName, \n\t\t$maxBetaLength, $lengthToUse, $correctFlag, $normParam \n";
					print MAINERR "\t\t$mySequence, $model \n";
				}
				
				singleReformatSwap($currentInName, $maxBetaLength, $lengthToUse, 
									$correctFlag, $normParam, $mySequence, $model);
				
				my $currentMatlabName = "$mySpecies $mybase $swapSuffix";
				unless (defined($opts->{'g'})) {

					my $rangeFlag;
					if ($correctFlag) { $rangeFlag = 2; }
					elsif ($mybase eq "prion single strand") { $rangeFlag = 1; }
					else { $rangeFlag = 0; }
					
					my $currentMatlabExecutable;
					$currentMatlabExecutable = "bsplotsingle400$idTime('$currentMatlabName', $maxBetaLength, $rangeFlag, '$optString', '$imageOpt', $countShift)";
					
					print STDERR "...";
					print MAINERR "\tCalling Matlab: $currentMatlabExecutable \n";
					my $matlabOutput = $matlab->execute($currentMatlabExecutable);
					print MAINERR "\t\tResult: $matlabOutput \n";
				} # end unless
				print STDERR "\n";
			} # end foreach $mybase
			if ($timeFlag) {
				$timeRef{"03.2.2 singleton $mySpecies"} = time - $runTime;
			}
		} # end if genFlag != 2
	} # end if -s is defined

### 3.3: Calculate and plot local maxima ###
	
	unless (defined($opts->{'f'})) {
		unless ($genFlag == 1) {
			my @pairPrefixes = ("$mySpecies 3dranged");
			if (defined $opts->{'A'}) {
				push (@pairPrefixes, "$mySpecies 3dranged antiparallel");
			}
			foreach my $pairPrefix (@pairPrefixes) {
				print MAINERR "\tCalling pairMaximaFinder: $pairPrefix \n";
				pairMaximaFinder($pairPrefix);
				if ($timeFlag) {
					$timeRef{"03.3.1 pairMaxima $pairPrefix"} = time - $runTime;
				} # end if
			} # end foreach
		} # end unless

		unless ($genFlag == 2) {
			my $singlePrefix = "$mySpecies prion single strand ";
			print MAINERR "\tCalling singleMaximaFinder: $singlePrefix \n";
			singleMaximaFinder($singlePrefix, $filter);
			if ($timeFlag) {
				$timeRef{"03.3.2 singleMaxima $mySpecies"} = time - $runTime;
			}
		}
	} # end unless

##################################################################
### 4. Organizing Output Files                                 ###
##################################################################
	
	# implementation of -d (creating directories and stuffing data and image files) here
	unless(defined($opts->{'d'})) {

### 4.1 List of directories to create, if needed ###
	
		print STDERR "4.\tSorting into directories\n\t";
		print MAINERR "Sorting into directories\n";
		
		my %topdirectories = (	"txt" => "*.txt",
								"diary" => "*.diary");
		if ($imageOpt eq "all") {
			$topdirectories{"jpg"} = "*.jpg";
			$topdirectories{"eps"} = "*.eps";
			$topdirectories{"tiff"} = "*.tiff";
			$topdirectories{"pdf"} = "*.pdf";
		}
		else {
			$topdirectories{"$imageOpt"} = "*.$imageOpt";
		}
		my %middirectories = (	"swap" => "*swap*",
								"single" => "*single*strand*",
								"combined" => "*combined*scores*",
								"forward" => "*forward*scores*");
		my %lowdirectories = (	"out-in" => "*scores*out-in*",
								"in-out" => "*scores*in-out*");
								
### 4.2 Main result directory, time-stamped and species-labeled ###
												
		my $startDir = $cwd;
		my $timeStamp = sprintf("%05d", $idTime);
		my $dirTag;	
		if ($address ne 'DEADBEEF') {
			$dirTag =  $address . "/". $mySpecies . "-". $timeStamp;
		}
		else {
			$dirTag =  $mySpecies . "-" . $timeStamp;
		}
		my $mainDir = $cwd . "/" . $dirTag;
		mkdir ($mainDir) or print MAINERR "Bug: directory $mainDir cannot be created \n";
		my $mvIDCmd = "mv *-". $idTime . ".* \"" . $mainDir . "\"";
		if (system($mvIDCmd)) { print MAINERR "Bug: moving working files failed for command $mvIDCmd \n"; }
		$mvIDCmd = "mv *00". $idTime . ".* \"" . $mainDir . "\"";
		if (system($mvIDCmd)) { print MAINERR "Bug: moving working files failed for command $mvIDCmd \n"; }
		foreach my $mytag (keys %topdirectories) {
# 			print $cwd . "\n";
			print STDERR "[$mytag]";
			opendir (CWD, $cwd);
			my @mainfileList = readdir(CWD);
			closedir (CWD);
			my $mainFile = 0;
			foreach my $filename (@mainfileList) {
				if ($filename =~ /^$mySpecies/ && $filename =~ /.$topdirectories{$mytag}/) { $mainFile++; }
			}
			if ($mainFile != 0) {
				my $mvCmd = "mv " . '"' . $mySpecies . "\"" . $topdirectories{$mytag} ." \"" . $dirTag . "\"";
#				print $mvCmd;
#				print STDERR "\n";
#				if (-d $dirTag) {
					if (system($mvCmd)) { print MAINERR "Bug: main move failed for $mytag\n"; }
#				}
			}
			else {
				print MAINERR "No files of type $mytag found, not moved from working directory \n";
			}
		} # end foreach my $mytag
		
### 4.3 Sorting and subdividing of files into subdirectories ###
		
		chdir($mainDir);
		my @fileList;

### 4.3.1 Primary subdivision by file extension ###
			
		foreach my $mydirectory (keys %topdirectories) {
			opendir (MAINDIR, $mainDir);
			@fileList = readdir(MAINDIR);
			closedir (MAINDIR);
			my $topFile = 0;
			foreach my $filename (@fileList) {
				if ($filename =~ /.$topdirectories{$mydirectory}/) { $topFile ++; }
			}
			if ($topFile != 0) {
				print STDERR "{$mydirectory";
				my $topDir = $mainDir . "/" . $mydirectory;
				mkdir ($topDir) or print MAINERR "Bug: directory $topDir cannot be created \n";
				my $mvCmd = "mv ". $topdirectories{$mydirectory} . " " . $mydirectory;
				system ("sync");
				if (system ($mvCmd)) { print MAINERR "Bug: move for $mydirectory failed \n"; }
				chdir ($topDir);

### 4.3.2 Secondary subdivision by calculation ###

				my @midKeys = keys %middirectories;
				@midKeys = sort @midKeys;
				foreach my $midDirectory (@midKeys) {
					opendir (TOPDIR, $topDir);
					@fileList = readdir(TOPDIR);
					closedir (TOPDIR);
					my $midFile = 0;
					my $searchPath = $middirectories{$midDirectory};
					$searchPath =~ s/\*/ /g;
					print MAINERR $searchPath . "\n";
					foreach my $filename (@fileList) {
						if ($filename =~ /$searchPath/) { $midFile ++; }
					}
					if ($midFile != 0) {
						print STDERR "|";
						my $midDir = $topDir . "/" . $midDirectory;
						mkdir ($midDir) or print MAINERR "Bug: directory $midDir cannot be created \n";
						$mvCmd = "mv ". $middirectories{$midDirectory} . " " . $midDirectory;
						system ("sync");
						if (system ($mvCmd)) { print MAINERR "Bug: move for $midDir failed \n"; }
						chdir ($midDir);
						
### 4.3.3 Tertiary subdivision by orientation, if necessary ###

						foreach my $lowDirectory (keys %lowdirectories) {
							opendir (MIDDIR, $midDir);
							@fileList = readdir(MIDDIR);
							closedir (MIDDIR);
							my $lowFile = 0;
							my $lowsearchPath = "$lowdirectories{$lowDirectory}";
							$lowsearchPath =~ s/\*/ /g;
							print MAINERR $lowsearchPath . "\n";
							foreach my $filename (@fileList) {
								if ($filename =~ /$lowsearchPath/) { $lowFile ++; }
							}
							if ($lowFile != 0) {
								print STDERR ".";
								my $lowDir = $midDir . "/" . $lowDirectory;
								mkdir ($lowDir) or print MAINERR "Bug: directory $lowDir cannot be created \n";
								$mvCmd = "mv ". $lowdirectories{$lowDirectory} . " " . $lowDirectory;
								system ("sync");
								if (system ($mvCmd)) { print MAINERR "Bug: move for $lowDir failed \n"; }
								
### 4.3.4 Escapes in case no such files exist to be sorted ###

							}	
							else {
								print MAINERR "No files of type $lowDirectory present, skipping \n";
							} # endif $lowFile != 0
						} # loops lowDirectory
						chdir ($topDir);
					} 
					else {
						print MAINERR "No files of type $midDirectory present, skipping \n";
					} #endif $midFile != 0
				} # loops midDirectory
				chdir ($mainDir);
				print STDERR "}";
			} 
			else {
				print MAINERR "No files of type $mydirectory present, skipping \n";
			} # endif $topFile ! = 0
		} # loops mydirectory
		chdir ($startDir);
		print STDERR "\n\tSorting complete\n";
		if ($timeFlag) {
			$timeRef{"04.4 directory sorting $mySpecies"} = time - $runTime;
		}

### 4.4 Generate zip file ###		
		
		if (defined($opts->{'Z'})) {
			my $tarFileName = "betascan-" . $timeStamp . ".tar.gz";
			my $zipCmd = "tar -z";
			if (-e $tarFileName) {
				$zipCmd = $zipCmd . "A";
			} 
			else {
				$zipCmd = $zipCmd . "c";
			} # end if -e $tarFileName
			$zipCmd = $zipCmd . " -f ". $timeStamp . " " . $dirTag . "/";
			system ($zipCmd);
			print STDERR "\n\tZip archive updated\n";
		} # end -Z code
	} # end -d code


##################################################################
### 5. Cleanup                                                 ###
##################################################################

### 5.1 End main loop from section 3.2 ###
	print MAINERR "Done with $mySpecies \n";
} # loops $mySpecies

### 5.2 Get rid of unwanted files ###

my $knownJunkCmd = "rm cmd\*.m";
system($knownJunkCmd);

print STDERR "5.\tCleaning up...\n";

if ($timeFlag) {
	$timeRef{"05.2 cleanup"} = time - $runTime;
}

##################################################################
### 6. Make web page                                           ###
##################################################################

if ($htmlFlag) {
	print STDERR "6.\tGenerating webfile...";
	makeHTML($cwd, $idTime, $htmlFileName, $timeString, $optString, 
				$inputFile, $minLoop, $maxLoop, $imageOpt, $model, 
				$weightFileName, $opts, $maxBetaLength, 
				$maxSeqLength, \@proteinNames);
	print STDERR "\n\tDone. Results available in $htmlFileName\n";
}
if (defined $opts->{'Z'}) {
	print STDERR "Results zipped in $tarFileName\n";
}

if ($timeFlag) {
	$timeRef{"06.2 webpage"} = time - $runTime;
	
	my @timeKeys = keys (%timeRef);
	@timeKeys = sort @timeKeys;
	print MAINERR "Time marks: \n"; 
	foreach my $timeKey (@timeKeys) {
		print MAINERR "\t $timeRef{$timeKey} \tseconds elapsed at or in $timeKey  \n";
	}
}

close MAINERR;

print STDERR "BETASCAN exiting normally.\n";

##################################################################
### END OF MAIN CODE                                           ###
##################################################################


##################################################################
### A. sub usage                                               ###
### purpose: should be obvious                                 ###
##################################################################

sub usage {
 print STDERR "usage: betascan.pl -n input.fasta \n";
 print STDERR "optional parameters: -b maxbetalength -l maxsequencelength -m minloop -M maxloop";
 print STDERR "                     -N normparam -i {all|jpg|pdf|tiff} -c model -j jobname\n";
 print STDERR "optional external file use: -w weightfilename -p probfilename \n";
 print STDERR "optional switches: -1 -2 -d -f -g -o -h -a -F -D -A -T\n";
 print MAINERR "Bad input parameters.\n";
 close MAINERR;
 exit;
}

##################################################################
### B. sub threeDRanger                                        ###
### purpose: rearrange tab-delimited file of form (x,y,z)      ###
### into tab-delimited file arranged as x vs y table           ###
##################################################################

# rearrange tab-delimited file of form (x,y,z) into table x vs y
sub threeDRanger {

### B.1 Read in passed variables ###
	my $inputName = shift;
	my $outputName = shift;
	my $length = shift;
	my $maxSeqLength = shift;
	my $sequence = shift;
	my $combinedFlag = shift;
	my $altsequence = "";
	if (defined $opts->{'a'}) {
		$altsequence = $seqHashRef->{$opts->{'a'}};
	}
	
	# for time reading
	my $startFxn = time;

### B.2 Read input file to hash ###

	unless (open(INFILE, $inputName)) {
		print STDERR "3dranger dies: Cannot open input file \"$inputName\"\n";
		exit;
	}
	print MAINERR "In threeDranger, file opened \n";
	
	my $minX = 10000;
	my $maxX = 0;
	my $minY = 10000;
	my $maxY = 0;
	my %dataHash;
	my $myline;
	
	foreach $myline (<INFILE>) {
		my $myX;
		my $myY;
		my $myZ;
		($myX, $myY, $myZ) = split('\t', $myline);
		chomp $myZ;
#		print MAINERR "myX = $myX, myY = $myY, myZ = $myZ, $maxX = $maxX, maxY = $maxY \n"; 
		if ($myX > $maxX) {$maxX = $myX;}
		if ($myX < $minX) {$minX = $myX;}
		if ($myY > $maxY) {$maxY = $myY;}
		if ($myY < $minY) {$minY = $myY;}
		my $keyName = join("by",($myX, $myY));
		$dataHash{$keyName} = $myZ;
	}

## B.3 Print data in rearranged form to output file ###

	unless (open(OUTFILE, ">$outputName")) {
		print STDERR "3dranger dies: Cannot open output file \"$outputName\"\n";
		exit;
	}

	# top row

	print OUTFILE "\t";
	for ( my $currentY = $minY; $currentY <= $maxY; ++$currentY) {
		if ($currentY < $maxSeqLength) {
			print OUTFILE $currentY."\t";
		}
	}
	print OUTFILE "\n";

	# main body

	for ( my $currentX = $minX; $currentX <= $maxX; ++$currentX) {
		if ($currentX < $maxSeqLength) {
			print OUTFILE $currentX."\t";
			for ( my $currentY = $minY; $currentY <= $maxY; ++$currentY) {
				if ($currentY < $maxSeqLength) {
					my $myKey = join("by", ($currentX, $currentY));
					if (defined $dataHash{$myKey}) {
						if ($dataHash{$myKey} != 0) {
								my $totalScore = $dataHash{$myKey};
								print OUTFILE $totalScore."\t";
						} 
						else {
							print OUTFILE "nan\t";
						} # end if ...!= 0
					} # end if defined....
					else
					{
						print OUTFILE "nan\t";
					}
				} # end if $currentY...
			} # end for my $currentY...
			print OUTFILE "\n";
		} # end if $currentX...
	} # end for my $currentX...

	close OUTFILE;

	print MAINERR "\t\tDone with threeDRanging. \n";
	if ($timeFlag) {
		if (exists $timeRef{"B threeDranger"}) {
			$timeRef{"B threeDranger"} += (time - $startFxn);
		}
		else {
			$timeRef{"B threeDranger"} = time - $startFxn;
		}
	} # end if $timeFlag
}

##################################################################
### C. sub singleReformatSwap                                  ###
### purpose: reformat singleton files as a single table        ###
### calculates and applies log-odds corrections                ###
##################################################################

# sub singleReformatSwap
# take beta-finding prion singleton files and reformat as one table
sub singleReformatSwap {

### C.1 Read in passed files and initialize variables ###

	my $prefix = shift;
	my $maxBetaLength = shift;
	my $maxSeqLength = shift;
	my $correctFlag = shift;
	my $normParam = shift;
	my $sequence = shift;
	my $model = shift;
	
	# for time
	my $startFxn = time;
	
	my $inInfix = "out-in";
	my $outInfix = "in-out"; 
	my $suffix = "residue.txt";
	my @data = ();
	my $isIn;
	my $length;
	my $myX;
	my $myY;
	
	print MAINERR "In singleReformatSwap \n";

### C.2 Read in data files and store ###

# 2*$maxBetaLength files are expected
# format "$prefix {in-out|out-in} {2|3|4...n-1|n} residue.txt

	for ($isIn = 0; $isIn <= 1; ++$isIn) {
	   for ($length = 2; $length <= $maxBetaLength; ++$length) {

### C.2.1 Read in data and store, with correction ###

		my $inputName = '';
		my $twoDigit = sprintf("%02d", $length);
		if ($isIn == 0) {
			$inputName = "$prefix $outInfix $twoDigit $suffix";
		}
		else {
			$inputName = "$prefix $inInfix $twoDigit $suffix";
		}
		unless (open(INFILE, $inputName)) {
	        	print STDERR "singleReformat dies: Cannot open input file";
				print STDERR "\"$inputName\"\n";
	        	exit;
		}
		my @inLine = <INFILE>;
		close INFILE;
		foreach my $myLine (@inLine) {
			($myX, $myY) = split('\t', $myLine);
			chomp $myY;
			my $rearrange = ($myX % 2);
			my $isRearranged = ($rearrange + $isIn) % 2;
			if ($myY == 0) {
				$data[$isRearranged][$length][$myX] = "nan";
			}
			else {
#				my $strandSeq = substr($sequence, $myX, $length);
#				my $correction = corrector($model, $strandSeq, 0);
				$data[$isRearranged][$length][$myX] = $myY;
#				 - $correction;
			}
		} # end foreach my $myLine
	  } # end for $length...
	} # end for $isIn...

### C.3 Print out rearranged and corrected data to file ###

	my $newSuffix = "complete swap.txt";
	my $outFileName = "$prefix $newSuffix";
	unless (open(OUTFILE, ">$outFileName")) {
		print STDERR "singleReformat dies: Cannot open output file \"$outFileName\"\n"; 
		exit;
	}

	# top row

	print OUTFILE "$outFileName\t";
	for ($isIn = 0; $isIn <= 1; ++$isIn) {
		for ($length = 2; $length <=$maxBetaLength; ++$length) {
			if ($isIn == 0) {print OUTFILE $outInfix;}
			else {print OUTFILE $inInfix;}
			print OUTFILE " $length\t";
		}	
	}
	print OUTFILE "\n";

	# main body

	for ($myX = 0; $myX < $maxSeqLength; ++$myX) {
		print OUTFILE "$myX\t";
		for ($isIn = 0; $isIn <= 1; ++$isIn) {
			for ($length = 2; $length <=$maxBetaLength; ++$length) {
#				print MAINERR "$isIn, $length, $myX \n";
				my $endOfBeta = $length + $myX;
				if(defined($data[$isIn][$length][$myX])) {
					if ($endOfBeta <= $maxSeqLength && $data[$isIn][$length][$myX] != 0) {
						print OUTFILE ($data[$isIn][$length][$myX]);
					}
					else { print OUTFILE "nan"; }
				}
				print OUTFILE "\t";
			} # end for $length...
		} # end for $isIn...
		print OUTFILE "\n";
	} # end for $myX...	
	close OUTFILE;
	
	print MAINERR "\t\tDone with singleReformatSwap \n";
	if ($timeFlag) {
		if (exists $timeRef{"C singleReformatSwap"}) {
			$timeRef{"C singleReformatSwap"} += (time - $startFxn);
		}
		else {
			$timeRef{"C singleReformatSwap"} = time - $startFxn;
		}
	} # end if $timeFlag

}

##################################################################
### D. sub singleMaximaFinder                                  ###
### purpose: calculate local maxima in singleton predictions   ###
### and invoke matlab to graph them                            ###
##################################################################

# singleMaximaFinder takes a single argument, the prefix for the complete swap file
# for singleton log-odds (calculated by singleReformatSwap).
# it outputs a tab-delimited file with the local maxima
# in the format position, length, orientation, score.
# note that a mod may be necessary to throw out trivials... once i know what trivials there are.
sub singleMaximaFinder {

### D.1 Read in passed variables and initialize ###
	# get the file
    my $singlePrefix = shift;
	my $filter = shift;	
	my $swapSuffix = "complete swap.txt";
	my $maxSuffix = "complete maxima.txt";
	my $singleName = $singlePrefix . $swapSuffix;
	
### D.2 Open and parse input file ###	
	unless (open (SINGLEFILE, $singleName)) { 
		print "Maxima finding aborts: can't find singleton file $singlePrefix\n"; 
		return;
	}
	# parse file
	# the first line defines everything
	my $singleHeader = <SINGLEFILE>;
	my @headers = split('\t', $singleHeader);
	my $numColumns = (scalar(@headers) - 2) / 2;
	
	print MAINERR "In singleMaximaFinder and file opened okay \n";
	print STDERR "3.3.1\tFinding single maxima.";
	# the rest is data
	my @singleFile = <SINGLEFILE>;
	close SINGLEFILE;
	my $seqLength = scalar(@singleFile);

	# dataLattice is a hash that is a more efficient multi-D array
	# the coordinates form the key
	# three dimensions here: position, length, orientation
	# keys are of form "P L O"
	# orientation 0 is odds buried, 1 is odds exposed
	my %dataLattice;
	foreach my $myLine (@singleFile) {
		my @lineData = split('\t', $myLine);
		my $counter = 0;
		for ($counter = 1; $counter <= $numColumns; $counter++) {
			my $length = $counter + 1;
			$dataLattice{"$lineData[0] $length 0"} = $lineData[$counter];
		}
		for ($counter = $numColumns + 1; $counter <= (2*$numColumns); $counter++) {
			my $length = $counter + 1 - $numColumns;
			$dataLattice{"$lineData[0] $length 1"} = $lineData[$counter];
		}
	} #end foreach my $myLine (@singleFile)

	
	# every self-referential member of this lattice is a local maximum. output those.
	# i calculate them all in case i care later.

### D.3 Begin output file ###

	my $outputFileName = $singlePrefix . $maxSuffix;
	print MAINERR "Output file will be $outputFileName\n";
	unless (open (SINGLEOUT, ">$outputFileName")) {
		die "Betascan dies in maxima finding mode: can't open singleton output file $outputFileName\n";
	}
	# Print out statistics to aid MATLAB
	print SINGLEOUT "Maxima for $headers[0]\n";
	print MAINERR "number of columns = $numColumns, sequence length = $seqLength\n";
	my $maxBeta = $numColumns + 1;
	print SINGLEOUT "$numColumns\t$seqLength\t0\t0\n";

### D.4 Process data ###

	# data are now read in. we need to process them.
	# the basic idea is to find each point's neighboring maximum
	# where the neighboring maximum is the maximum value in the set of points
	# (P,L,O), (P,L,[O+1 mod 2]), (P-1,L+1,O), (P+1,L-1,O), (P,L+1,O), (P,L-1,O).
	# I.e., possible motions in the space of singletons are:
	# Do nothing (trivial)
	# Flip orientation
	# Add or subtract a residue from the strand at the N-terminal end
	# Add or subtract a residue from the strand at the C-terminal end

### D.4.1 Initializations ###
	
	my %nextLattice;
	my %maxInLattice;
	my %maxOutLattice;
	my $counter;
	foreach my $point (keys %dataLattice) {
	
		# Keep the user happy
		$counter++;
		if ($counter == 500) {
			print STDERR ".";
			$counter = 0;
		}
	
		# if this point is undefined, forget about it.
		if ($dataLattice{$point} =~ /nan/) { next; }
		# assemble the set of neighbors.
		my @coords = split(' ',$point);
		my @neighbors;

### D.4.2 Check for existence of each neighbor ###		

		# (P,L,O): always exists
		$neighbors[0] = $point;
		# (P,L,[O+1 mod 2]: always exists
		push(@neighbors, join(' ',($coords[0], $coords[1], (($coords[2] + 1) % 2))));
		# (P-1,L+1,O): exists unless P = 0 or L = $numColumns+1
		unless ($coords[0] == 0 || $coords[1] == ($numColumns+1)) {
			push (@neighbors, join(' ',(($coords[0] - 1), ($coords[1] + 1), $coords[2])));
		}
		# (P+1,L-1,O): exists unless P = $seqLength or L = 2 
		unless ($coords[0] == $seqLength || $coords[1] == 2) {
			push (@neighbors, join(' ',(($coords[0] + 1), ($coords[1] - 1), $coords[2])));
		}
		# (P, L+1, O): exists unless P = ($seqLength - L + 1) or L = $numColumns + 1
		unless ($coords[0] == ($seqLength - $coords[1] - 1) || $coords[1] == ($numColumns+1)) {
			push (@neighbors, join(' ',($coords[0], ($coords[1] + 1), $coords[2])));
		}
		# (P, L-1, O): exists unless L = 2
		unless ($coords[1] == 2) {
			push (@neighbors, join(' ',($coords[0], ($coords[1] - 1), $coords[2])));
		}

### D.4.3 Find maximum of nearest neighbors; if maximum = self, we have a local maximum ###
		
		# all neighbors found.
		# now, find the maximum of the neighbors and put it in the next lattice.
		my $max = $point;
		foreach my $myneighbor (@neighbors) {
			if (!exists ($dataLattice{$point}) || !exists ($dataLattice{$myneighbor}) ||
				$dataLattice{$point} =~ /nan/ || $dataLattice{$myneighbor} =~ /nan/) {
				print MAINERR "point = $point with value "; 
				if (exists($dataLattice{$point})) {
					print MAINERR "$dataLattice{$point}; ";
				}
				else {
					print MAINERR "UNDEFINED!!! \n";
				}
				print MAINERR "neighbor = $myneighbor with value ";
				if (exists($dataLattice{$myneighbor})) {
					print MAINERR "$dataLattice{$myneighbor} \n";
					if ($dataLattice{$myneighbor} =~ /[0-9]+/ && $dataLattice{$point} =~ /[0-9]+/) {
						if ($dataLattice{$myneighbor} > $dataLattice{$point}) {
							$max = $myneighbor;
						}
					}
				}
				else {
					print MAINERR "UNDEFINED!!! \n";
				}
			}
			else {
				if ($dataLattice{$myneighbor} =~ /[0-9]+/ && $dataLattice{$point} =~ /[0-9]+/) {
					if ($dataLattice{$myneighbor} > $dataLattice{$point}) {
						$max = $myneighbor;
					}
				}
			}
		}
		$nextLattice{$point} = $max;
		# if the two are identical this is a local maximum so ready it for output.
		if ($max eq $point) {
			# now we get cute. in $maxLattice, the DATA are the keys and the POSITIONS are the values.
			if ($coords[2] == 1) {
				$maxOutLattice{$dataLattice{$point}} = $point;
			}
			else {
				$maxInLattice{$dataLattice{$point}} = $point;
			}
#			push(@coords,$dataLattice{$point});
#			foreach my $co (@coords) {
#				print SINGLEOUT $co . "\t";
#			}
#			print SINGLEOUT "\n";
		}
	} # end foreach my $point (keys %dataLattice) 

### D.5 Sort and caculate statistics of local minima ###
	
	# we have moved the maxima to their own in-out and out-in lattices.
	# we sort each by score, and output.
	my @inScores = keys %maxInLattice;
	my @outScores = keys %maxOutLattice;
	@inScores = sort {$a <=> $b} @inScores;
	@outScores = sort {$a <=> $b} @outScores;
	my @inRScores = reverse @inScores;
	my @outRScores = reverse @outScores;
	
	my $numInScores = scalar(@inRScores);
	my $numOutScores = scalar (@outRScores);
	my $numInPosScores = 0;
	my $numOutPosScores = 0;

### D.6 Output local minima and statistics ###
	
	foreach my $inNum (@inRScores) {
		my @inNumPos = split(' ',$maxInLattice{$inNum});
		push (@inNumPos, $inNum);
		foreach my $co (@inNumPos) {
			print SINGLEOUT $co . "\t";
		}
		print SINGLEOUT "\n";
		if ($inNum > 0) {
			$numInPosScores ++;
		}
	}
	foreach my $outNum (@outRScores) {
		my @outNumPos = split(' ',$maxOutLattice{$outNum});
		push (@outNumPos, $outNum);
		foreach my $co (@outNumPos) {
			print SINGLEOUT $co . "\t";
		}
		print SINGLEOUT "\n";
		if ($outNum > 0) {
			$numOutPosScores ++;
		}
	}
	
	print SINGLEOUT "$numInScores\t$numInPosScores\t$numOutScores\t$numOutPosScores\n";
	close SINGLEOUT;
	
### D.7 Render matlab plot ###	
	
	unless (defined $opts->{'g'}) {
		print STDERR "Rendering...";
		my $matlabSinglePlot = "bsbetasingle400$idTime ('$outputFileName', '$optString', $countShift)";
		print MAINERR "\tCalling Matlab: $matlabSinglePlot \n";
		my $matlabOutput = $matlab->execute($matlabSinglePlot);
		print MAINERR "\t\tResult: $matlabOutput \n";
		print STDERR "\n";
	}

}

##################################################################
### E. sub pairMaximaFinder                                    ###
### purpose: calculate local maxima in pairwise predictions    ###
### and invoke matlab to graph them                            ###
##################################################################

#pairMaximaFinder finds the local maxima of a set of pairwise combined files.
#it takes the prefix as input
#and outputs a tab-delimited file with a line for each maximum
#with the positions, lengths, orientations, and score.
#may require modding to eliminate trivials.
#assumes all files are in one directory -- may need changing later.
sub pairMaximaFinder {

### E.1 Initialize and grab passed variables ###

	my $pairPrefix = shift;
	$pairPrefix = $pairPrefix . " prion combined scores";
	my $maxSuffix = " complete maxima.txt";
	my $Columns = 0;
	my $Rows = 0;
	my $antiParallel = 0;
	if (defined $opts->{'A'} && ($pairPrefix =~ /antiparallel/)) {
		$antiParallel = 1;
	}
	
	# a hash is far more efficient and easy to work with than an array
	# hash keys are of the form
	# "P1 L1 O1 P2 L2 O2"
	# strand 1 has positions across (columns)
	# strand 2 has positions down (rows)
	my %pairLattice;
	
	print MAINERR "In pairMaximaFinder \n";

### E.2 Read in data from files ###
	
	foreach my $myOrient (@orient) {
	
### E.2.1 Grab the correct file ###	

		my $orientValue;
		if ($myOrient =~ /in-out/) {
			$orientValue = 0;
		}
		elsif ($myOrient =~ /out-in/) {
			$orientValue = 1;
		}
		else {
			print MAINERR "Orientation has serious problems. \n";
		}
		foreach my $myLength (@length) {
			my $twoDigit = sprintf("%02d", $myLength);
			my $currentName = "$pairPrefix $myOrient $twoDigit $suffix";
			unless (open (PAIRFILE, $currentName)) {
				print "Pair maxima finding aborts: can't open file $currentName"; 
				return;
			}
			
### E.2.2 Grab data from file ###
			
			# parse file
			# the first line defines everything
			my $pairHeader = <PAIRFILE>;
			my @headers = split('\t', $pairHeader);
			my $numColumns = (scalar(@headers) - 1);
			$Columns = $numColumns;
#			print MAINERR "columns = $Columns\n";
			
			# the rest is data
			my @pairFile = <PAIRFILE>;
			close PAIRFILE;
			my $numRows = scalar(@pairFile);
			$Rows = $numRows;
#			print MAINERR "rows = $Rows\n";
			
### E.2.3 Parse the file data into the lattice ###

			foreach my $myLine (@pairFile) {
				my @lineData = split('\t', $myLine);
				my $counter = 0;
				for ($counter = 1; $counter <= $numColumns; $counter++) {
					my $position = $counter - 1;
					unless ($lineData[$counter] =~ /\s/) {
						$pairLattice{"$position $myLength $orientValue $lineData[0] $myLength $orientValue"} = $lineData[$counter];
					}
				}
			} #end foreach my $myLine (@pairFile)
		} # end foreach my $myLength
	} # end foreach my $myOrient
	
	# We have now filled the data lattice. Now we must calculate the next lattice.
	# the basic idea is to find each point's neighboring maximum
	# where the neighboring maximum is the maximum value in the set of points
	# possible motions in the space of pairwises are:
	# (P1,		L1,		O1,				P2,		L2,		O2				): Do nothing (trivial)
	# (P1,		L1,		[O1+1 mod 2],	P2,		L2,		[O2+1 mod 2]	): Flip orientation
	# (P1,		L1+1,	O1,				P2,		L2+1,	O2				): Add a residue from the strand at the C-terminal end
	# (P1,		L1-1,	O1,				P2,		L2-1,	O2				): Subtract a residue from the strand at the C-terminal end
	# (P1-1,	L1+1,	O1,				P2-1,	L2+1,	O2				): Add a residue from the strand at the N-terminal end
	# (P1+1,	L1-1,	O1,				P2+1,	L2-1,	O2				): Subtract a residue from the strand at the N-terminal end
	# (P1,		L1,		O1,				P2-1,	L2,		O2				): Slide relative connection 1 residue, moving lower strand left
	# (P1,		L1,		O1,				P2+1,	L2,		O2				): Slide relative connection 1 residue, moving lower strand right
	# (P1-1,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 1 residue, moving upper strand left
	# (P1+1,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 1 residue, moving upper strand right
	# (P1,		L1,		O1,				P2-2,	L2,		O2				): Slide relative connection 2 residue, moving lower strand left
	# (P1,		L1,		O1,				P2+2,	L2,		O2				): Slide relative connection 2 residue, moving lower strand right
	# (P1-2,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 2 residue, moving upper strand left
	# (P1+2,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 2 residue, moving upper strand right
	
	# and if that weren't enough, some of them change for antiparallel!
	
	# possible motions in the space of ANTIPARALLEL pairwises are:
	# (P1,		L1,		O1,				P2,		L2,		O2				): Do nothing (trivial)
	# (P1,		L1,		[O1+1 mod 2],	P2,		L2,		[O2+1 mod 2]	): Flip orientation
	##(P1,		L1+1,	O1,				P2-1,	L2+1,	O2				): Add a residue from the strand at the C-terminal end
	##(P1,		L1-1,	O1,				P2-1,	L2-1,	O2				): Subtract a residue from the strand at the C-terminal end
	##(P1-1,	L1+1,	O1,				P2,		L2+1,	O2				): Add a residue from the strand at the N-terminal end
	##(P1+1,	L1-1,	O1,				P2,		L2-1,	O2				): Subtract a residue from the strand at the N-terminal end
	# (P1,		L1,		O1,				P2-1,	L2,		O2				): Slide relative connection 1 residue, moving lower strand left
	# (P1,		L1,		O1,				P2+1,	L2,		O2				): Slide relative connection 1 residue, moving lower strand right
	# (P1-1,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 1 residue, moving upper strand left
	# (P1+1,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 1 residue, moving upper strand right
	# (P1,		L1,		O1,				P2-2,	L2,		O2				): Slide relative connection 2 residue, moving lower strand left
	# (P1,		L1,		O1,				P2+2,	L2,		O2				): Slide relative connection 2 residue, moving lower strand right
	# (P1-2,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 2 residue, moving upper strand left
	# (P1+2,	L1,		O1,				P2,		L2,		O2				): Slide relative connection 2 residue, moving upper strand right
	
	# every self-referential member of this lattice is a local maximum. output those.
	# i calculate them all in case i care later.

### E.3 Initialize output file(s) with headers ###
	
	my $outputFileName = $pairPrefix . $maxSuffix;
	open (PAIROUT, ">$outputFileName") or die "Betascan dies in maxima finding mode: can't open pairwise output file $outputFileName\n";
	print STDERR "3.3.2\tFinding pairwise maxima.";
	print PAIROUT "$Columns\t$Rows\t0\t0\t0\t0\t0\n";
	
	my $errorFileName = $pairPrefix . " error " . $maxSuffix;
	open (PAIRERR, ">$errorFileName") or die "Betascan dies in maxima finding mode: can't open pairwise output file $errorFileName\n";

### E.4 Process data: find maximum of nearest neighbors ###
	
	my $nextLattice= {};
	my %maxInLattice;
	my %maxOutLattice;
	my $counter = 0;
	foreach my $point (keys %pairLattice) {
	
		# Keep the user happy
		$counter ++;
		if ($counter == 25000) {
			print STDERR ".";
			$counter = 0;
		}

### E.4.1 Initializations ###

		# if this point is undefined, forget about it.
		if ($pairLattice{$point} =~ /nan/) { next; }
		# assemble the set of neighbors.
		my @coords = split(' ',$point);

### E.4.2 Find all neighbors which exist ###

		my @neighbors =();
		# (P1,		L1,		O1,				P2,		L2,		O2				):					Do nothing (trivial)
		# Always exists
		$neighbors[0] = $point;
		# (P1,		L1,		[O1+1 mod 2],	P2,		L2,		[O2+1 mod 2]	):					Flip orientation
		# Always exists
		push(@neighbors, join(' ',($coords[0], $coords[1], (($coords[2] + 1) % 2), $coords[3], $coords[4], (($coords[5] + 1) % 2))));
		# (P1,		L1+1,	O1,				P2,		L2+1,	O2				): Parallel:		Add a residue from the strand at the C-terminal end
		# exists unless P1 = $Columns - L1 - 1 or P2 = $Rows - L1 - 1 or L1 = L2 = $maxBetaLength
		unless ($antiParallel) {
			unless ($coords[0] == ($Columns - $coords[1] - 1) || 
				$coords[3] == ($Rows - $coords[4] - 1) || 
				$coords[1] == $maxBetaLength || 
				$coords[4] == $maxBetaLength) {
				push (@neighbors, join(' ',($coords[0], $coords[1]+1, $coords[2], $coords[3], $coords[4]+1, $coords[5])));
			}
		}
		# (P1,		L1+1,	O1,				P2-1,	L2+1,	O2				): Antiparallel:	Add a residue from the strand at the C-terminal end
		# exists unless P1 = $Columns - L1 -1 or P2 = 0 or P2 = $Rows - L2 - 1 or L1 = L2 = $maxBetaLength
		else {
			unless ($coords[0] == ($Columns - $coords[1] - 1) ||
				$coords[3] == 0 ||
				$coords[3] == ($Rows - $coords[4] - 1) ||
				$coords[1] == $maxBetaLength ||
				$coords[4] == $maxBetaLength) {
				push (@neighbors, join(' ',($coords[0], $coords[1]+1, $coords[2], $coords[3]-1, $coords[4]+1, $coords[5])));
			}
		}
		# (P1,		L1-1,	O1,				P2,		L2-1,	O2				): Parallel:		Subtract a residue from the strand at the C-terminal end
		# Exists unless L1 = L2 = 2
		unless ($antiParallel) {
			unless ($coords[1] == 2 || 
					$coords[4] == 2) {
				push (@neighbors, join(' ',($coords[0], $coords[1]-1, $coords[2], $coords[3], $coords[4]-1, $coords[5])));
			}
		}
		# (P1,		L1-1,	O1,				P2+1,	L2-1,	O2				): Antiparallel:	Subtract a residue from the strand at the C-terminal end
		# Exists unless L1 = L2 = 2
		else {
			unless ($coords[1] == 2 ||
					$coords[4] == 2) {
				push (@neighbors, join(' ',($coords[0], $coords[1]-1, $coords[2], $coords[3]+1, $coords[4]-1, $coords[5])));
			}
		}
		# (P1-1,	L1+1,	O1,				P2-1,	L2+1,	O2				): Parallel:		Add a residue from the strand at the N-terminal end
		# Exists unless P1 = 0 or P2 = 0 or L1 = L2 = $maxBetaLength
		unless ($antiParallel) {
			unless ($coords[0] == 0 || 
				$coords[3] == 0 ||
				$coords[1] == $maxBetaLength ||
				$coords[4] == $maxBetaLength) {
				push (@neighbors, join(' ',($coords[0]-1, $coords[1]+1, $coords[2], $coords[3]-1, $coords[4]+1, $coords[5])));
			}
		}
		# (P1-1,	L1+1,	O1,				P2,		L2+1,	O2				): Antiparallel:	Add a residue from the strand at the N-terminal end
		# Exists unless P1 = 0 or P2 = $Rows - L2 -1 or L1 = L2 = $maxBetaLength
		else {
			unless ($coords[0] == 0 ||
					$coords[3] == ($Rows - $coords[4] - 1) ||
					$coords[1] == $maxBetaLength ||
					$coords[4] == $maxBetaLength) {
				push (@neighbors, join(' ',($coords[0]-1, $coords[1]+1, $coords[2], $coords[3], $coords[4]+1, $coords[5])));
			}		
		}
		# (P1+1,	L1-1,	O1,				P2+1,	L2-1,	O2				): Parallel:		Subtract a residue from the strand at the N-terminal end
		# Exists unless P1 = $Columns - L1 - 1 or P2 = $Rows - L1 - 1 or L1 = L2 = 2
		unless ($antiParallel) {
			unless ($coords[0] == ($Columns - $coords[1] - 1) || 
				$coords[3] == ($Rows - $coords[1] - 1) ||
				$coords[1] == 2 ||
				$coords[4] == 2) {
				push (@neighbors, join(' ',($coords[0]+1, $coords[1]-1, $coords[2], $coords[3]+1, $coords[4]-1, $coords[5])));
			}
		}
		# (P1+1,	L1-1,	O1,				P2,		L2-1,	O2				): Antiparallel:	Subtract a residue from the strand at the N-terminal end
		# Exists unless P1 = $Columns - L1 - 1 or L1 = L2 = 2		
		else {
			unless ($coords[0] == ($Columns - $coords[1] - 1) ||
					$coords[1] == 2 ||
					$coords[4] == 2 ) {
				push (@neighbors, join(' ',($coords[0]+1, $coords[1]-1, $coords[2], $coords[3], $coords[4]-1, $coords[5])));
			}
		}
		# (P1,		L1,		O1,				P2-1,	L2,		O2				):					Slide relative connection 1 residue, moving lower strand left
		# Exists unless P2 = 0
		unless ($coords[3] == 0) {
			push (@neighbors, join(' ',($coords[0], $coords[1], $coords[2], $coords[3]-1, $coords[4], $coords[5])));
		}
		# (P1,		L1,		O1,				P2+1,	L2,		O2				):					Slide relative connection 1 residue, moving lower strand right
		# Exists unless P2 = $Rows - L1 - 1
		unless ($coords[3] == $Rows - $coords[4] - 1) {
			push (@neighbors, join(' ',($coords[0], $coords[1], $coords[2], $coords[3]+1, $coords[4], $coords[5])));
		}
		# (P1-1,	L1,		O1,				P2,		L2,		O2				):					Slide relative connection 1 residue, moving upper strand left
		# Exists unless P1 = 0
		unless ($coords[0] == 0) {
			push (@neighbors, join(' ',($coords[0]-1, $coords[1], $coords[2], $coords[3], $coords[4], $coords[5])));
		}
		# (P1+1,	L1,		O1,				P2,		L2,		O2				):					Slide relative connection 1 residue, moving upper strand right
		# Exists unless P1 = $Columns - L1 - 1
		unless ($coords[0] == $Columns - $coords[1] - 1) {
			push (@neighbors, join(' ',($coords[0]+1, $coords[1], $coords[2], $coords[3], $coords[4], $coords[5])));
		}
		# (P1,		L1,		O1,				P2-2,	L2,		O2				):					Slide relative connection 2 residue, moving lower strand left
		# Exists unless P2 = 0 or 1
		unless ($coords[3] <= 1) {
#			print PAIRERR "SLIDE LOWER STRAND 2 LEFT: P2 = $coords[3], should be greater than 1\n";
			push (@neighbors, join(' ',($coords[0], $coords[1], $coords[2], $coords[3]-2, $coords[4], $coords[5])));
		}
		else {
#			print PAIRERR "Lower strand can't slide 2 left because $coords[3] is less than 1\n";
		}
		# (P1,		L1,		O1,				P2+2,	L2,		O2				):					Slide relative connection 2 residue, moving lower strand right
		# Exists unless P2 = $Rows - L1 
		unless ($coords[3] == $Rows - $coords[4] ) {
#			print PAIRERR "SLIDE LOWER STRAND 2 RIGHT: P2 = $coords[3], should be less than ".($Rows - $coords[4])."\n";
			push (@neighbors, join(' ',($coords[0], $coords[1], $coords[2], $coords[3]+2, $coords[4], $coords[5])));
		}
		else {
#			print PAIRERR "Lower strand can't slide 2 right because $coords[3] is too high\n";
		}
		# (P1-2,	L1,		O1,				P2,		L2,		O2				):					Slide relative connection 2 residue, moving upper strand left
		# Exists unless P1 = 0 or 1
		unless ($coords[0] <= 1) {
#			print PAIRERR "SLIDE UPPER STRAND 2 LEFT: P1 = $coords[0], should be greater than 1\n";
			push (@neighbors, join(' ',($coords[0]-2, $coords[1], $coords[2], $coords[3], $coords[4], $coords[5])));
		}
		else {
#			print PAIRERR "Upper strand can't slide 2 left because $coords[0] is less than 1.\n";
		}
		# (P1+2,	L1,		O1,				P2,		L2,		O2				):					Slide relative connection 2 residue, moving upper strand right
		# Exists unless P1 = $Columns - L1 
		unless ($coords[0] > $Columns - $coords[1]) {
#			print PAIRERR "SLIDE UPPER STRAND 2 RIGHT: P1 = $coords[0], should be less than " . ($Columns - $coords[1]) . "\n";
			push (@neighbors, join(' ',($coords[0]+2, $coords[1], $coords[2], $coords[3], $coords[4], $coords[5])));
		}
		else {
#			print PAIRERR "Upper strand can't slide 2 right because $coords[0] is too high\n";
		}

### E.4.3 Find maximum of nearest neighbors; if maximum = self, we have a local maximum ###

		# all neighbors found.
		# now, find the maximum of the neighbors and put it in the next lattice.
		my $max = $point;
		
		foreach my $myneighbor (@neighbors) {
			if (!exists ($pairLattice{$point}) || !exists ($pairLattice{$myneighbor})) {
				print PAIRERR "point = $point with value"; 
				if (exists($pairLattice{$point})) {
					print PAIRERR "$pairLattice{$point}; ";
				}
				else {
					print PAIRERR "UNDEFINED!!! \n";
				}
				print PAIRERR "neighbor = $myneighbor with value ";
				if (exists($pairLattice{$myneighbor})) {
					print PAIRERR "$pairLattice{$myneighbor} \n";
					if ($pairLattice{$myneighbor} > $pairLattice{$point}) {
						$max = $myneighbor;
					}
				}
				else {
					print PAIRERR "UNDEFINED!!! \n";
				}
			}
			elsif (!($pairLattice{$myneighbor} =~ /nan/)) {
				if ($pairLattice{$myneighbor} > $pairLattice{$point}) {
					$max = $myneighbor;
				}
			}
		}
		
		$nextLattice->{$point} = $max;
		
		# if the two are identical this is a local maximum so output.
		if ($max eq $point) {
		    if ($coords[2] == 1) {
				$maxOutLattice{$pairLattice{$point}} = $point;
			}
			else {
				$maxInLattice{$pairLattice{$point}} = $point;
			}
#			push(@coords,$pairLattice{$point});
#			foreach my $co (@coords) {
#				print PAIROUT $co . "\t";
#			}
#			print PAIROUT "\n";
		}
	} # end foreach my $point (keys %pairLattice) 
	
### E.5 Calculate fields of the maxima as attractors ###

	# nextLattice contains the links between points
	# zoneLattice goes from the maxima to a hash of items in the zone and their scores
	# reverseZoneLattice goes from each point to its maximum
	
	my $zoneLattice = {};
	my $reverseZoneLattice = {};
	foreach my $nextpoint (keys %$nextLattice) {
		my $attractor = findAttractor($nextpoint, $nextLattice, $reverseZoneLattice);
		$reverseZoneLattice->{$nextpoint} = $attractor;
		if (exists $zoneLattice->{$attractor}) {
			$zoneLattice->{$attractor}->{$nextpoint} = $pairLattice{$nextpoint};
		}
		else {
			$zoneLattice->{$attractor} = { $nextpoint => $pairLattice{$nextpoint} };
		}
	}
	
	# now that we have calculated the zoneLattice set, we can derive statistics from it
	# the number of elements in the array so created is the VOLUME of the zone
	
	my $zoneVolume = {};
	foreach my $zoneAttractor (keys %$zoneLattice) {
		$zoneVolume->{$zoneAttractor} = scalar($zoneLattice->{$zoneAttractor});
	}
	
	# the elements not surrounded by other members are on the BOUNDARY of the zone
	# we can talk about their number or their scores -- say their average

### E.6 Sort and caculate statistics of local maxima ###
	
	my @inScores = keys %maxInLattice;
	my @outScores = keys %maxOutLattice;
	@inScores = sort {$a <=> $b} @inScores;
	@outScores = sort {$a <=> $b} @outScores;
	my @inRScores = @inScores;
	my @outRScores = @outScores;
	
	my $numInScores = scalar(@inRScores);
	my $numOutScores = scalar (@outRScores);
	my $numInPosScores = 0;
	my $numOutPosScores = 0;

### E.7 Output local minima and statistics ###
	
	foreach my $inNum (@inRScores) {
		my @inNumPos = split(' ',$maxInLattice{$inNum});
		push (@inNumPos, $inNum);
		foreach my $co (@inNumPos) {
			print PAIROUT $co . "\t";
		}
		print PAIROUT "\n";
		if ($inNum > 0) {
			$numInPosScores ++;
		}
	}
	foreach my $outNum (@outRScores) {
		my @outNumPos = split(' ',$maxOutLattice{$outNum});
		push (@outNumPos, $outNum);
		foreach my $co (@outNumPos) {
			print PAIROUT $co . "\t";
		}
		print PAIROUT "\n";
		if ($outNum > 0) {
			$numOutPosScores ++;
		}
	}
	
	print PAIROUT "$numInScores\t$numInPosScores\t$numOutScores\t$numOutPosScores\t$antiParallel\t0\t0\n";
	close PAIROUT;
	close PAIRERR;	

### E.8 Render matlab plot ###
	unless (defined $opts->{'g'}) {
		print STDERR "Rendering...";
		my $matlabPairPlot = "bsbetapair400$idTime ('$outputFileName', '$optString', $countShift)";
		$matlab->execute($matlabPairPlot);
		print STDERR "\n";
	}

}

##################################################################
### F. sub corrector                                           ###
### purpose: calculate the null score for this strand          ###
### given its sequence and the model to be used                ###
### corrector("none", $sequence) will just give correction     ###
### by length and (if desired) by orientation                  ###
##################################################################

sub corrector {

	my $model = shift;
	my $sequence = shift;
	my $antiParallel = shift;
	
	# antiParallel correction uses direct numbers
	# from Penel et al., 2003 histogram.
	my @antiParallelHistogram = ( 0, 0,
		-1.4655355,
		-1.6925929,
		-1.9956430,
		-2.1586826,
		-2.2609615,
		-2.5693697,
		-2.7527389,
		-3.4149239,
		-3.8898175,
		-4.2952826,
		-4.7771207,
		-5.3367365,
		-5.5190580,
		-6.7230308,
		-7.1284959,
		-7.1284959,
		-6.7230308,
		-7.1284959 );
	
	my $startFxn = time;
	my $length = length $sequence;
	my $correction = 0;

	# correction for sequence
	# depends on the assumption we are making that residue identities are independent
	if ($model eq "none") {
		if (defined $opts->{'D'}) { print MAINERR "None "; }
		$correction = 0;
	}
	elsif ($model eq "null") {
		if (defined $opts->{'D'}) { print MAINERR "Null "; }
		$correction = $length * log(0.05);
	}
	else {
		# may be faster to read once, in main
		# my $weights = readWeightFile($weightFileName);
		# use the correct model
		# options: allproteins, allbetas, allprions, yeastproteins, yeastbetas, yeastprions
		# make sure this model exists!
		unless (exists($weights->{'dist'}->{$model}->{'A'})) {
			die "Betascan dies in corrector: can't find model $model in BSweights.txt!\n";
		}
		# calculate the sequence-based model for this sequence
		my @residues = split ('',$sequence);
		foreach my $currentResidue (@residues) {
			# $weights->{'dist'}->{$model}->{$currentResidue} = value to be frobbed.
			$correction += log($weights->{'dist'}->{$model}->{$currentResidue});
		}
	}
	# correction for length: frobbable with -L
	unless (defined $opts->{'L'}) {
		# parallel :Poisson model
		unless ($antiParallel) {
			$correction += $normParam - (($length - 1 )* (log($normParam)));
			for ($kk = ($length - 1); $kk > 1; $kk--) {
				$correction += log($kk);
			}
		}
	# antiparallel version: direct histogram
		else {
			$correction -= $antiParallelHistogram[$length]; 
		}
	}
	# correction for orientation
	if ($orientFlag == 1) {
		$correction += $length * log(2);
	}
	
	if (defined $opts->{'D'}) { 
		print MAINERR "Corrector: Corr = $correction using model $model for sequence $sequence \n";
	}
	
	if ($timeFlag) {
		if (exists $timeRef{"F corrector"}) {
			$timeRef{"F corrector"} += (time - $startFxn);
		}
		else {
			$timeRef{"F corrector"} = time - $startFxn;
		}
	} # end if $timeFlag

	return $correction;
}

##################################################################
### G. sub readWeightFile                                      ###
### purpose: pulls all null models out of the accompanying file###
### so that the corrector subroutine can use them              ###
### modded from code by Andrew McDonnell 6/2006 for OKOLONA    ###
##################################################################

sub readWeightFile {
 my $fname = shift;

 # positions, probability distributions
 my $weights = {'pos' => {}, 'dist' => {}}; 
 open F, $fname or die "Betascan dies in readWeightFile: can't open sequence model file $fname\n";
 my $dist = '';
 my $mode = 0;
 my $idx = 0;
 while (<F>) {
   next if $_ =~ /^\s*#/; # comment
   chomp;
   next if $_ =~ /^\s*$/;
   if ($_ =~ s/^WEIGHTNAME\s+//) { # a new weight distribution
     $dist = $_;
     $mode = 1;
   }
   elsif ($mode == 1 and $_ =~ /^([A-Z])\s+([0-9\.]+)/) {
     $weights->{'dist'}->{$dist}->{$1} = $2;
   }
# The WEIGHTASSIGN functionality allows blending of models in specific residues.
# This is useful in HMM weighting, for instance.
# We don't want it in Betascan, though: for comparability, all data must be 
# calculated 100% by the same model.   

#   elsif ($_ =~ s/^WEIGHTASSIGN\s+//) {
#     $idx = $_;
#     $mode = 2;
#   }
#   elsif ($mode == 2 and $_ !~ /^WEI/) {
#     my @l = split;
#     if (!defined $weights->{'dist'}->{$l[0]}) {
#	die "BETASCAN dies in readWeightFile: undefined distribution '$l[0]' in line '$_'";
#     }
#     $weights->{'pos'}->{$idx}->{$l[0]} = $l[1];
#   }
}
close F;

# sanity checking - do weights sum to 1, etc?
foreach my $d (keys %{$weights->{'dist'}}) {
	my $sum = 0;
	foreach my $aa (keys %{$weights->{'dist'}->{$d}}) {
		$sum += $weights->{'dist'}->{$d}->{$aa};
	}
	if (abs(1-$sum) > 0.001) { 
		die "Betascan dies in readWeightFile: distribution $d does not sum to 1 but $sum\n"; 
	}
 }

 return $weights;
}

##################################################################
### H. sub calcProbs                                           ###
### purpose: primary calculations for Betascan                 ###
##################################################################

sub calcProbs {

### H.1 Grab passed variables ###
	my $seqHashRef = shift;
	my $maxBetaLength = shift;
	my $minLoop = shift;
	my $maxLoop = shift;
	my $normParam = shift;
	my $model = shift;

### H.2 Initializations ###

	# These three structures define the BETAWRAP probability tables.
	# %gLookupTable binds one-letter codes to row/column assignments.
	# condProbIn and condProbOut are the actual tables,
	# all in log format (i.e., values given are ln(Pr(situation)).)

### H.2.1 Defaults ###
	
	# From the original BETAWRAP, 2003
	# Excludes all beta-helices (three-stranded right-handed) known at that time.
	
	my %gLookupTable = (
						'A' => 5,		'a' => 5,
						'B' => 20,		'b' => 20,
						'C'	=> 10,		'c' => 10,
						'D' => 16,		'd'	=> 16,
						'E'	=> 17,		'e' => 17,
						'F'	=> 2,		'f' => 2,
						'G' => 9,		'g' => 9,
						'H' => 8,		'h' => 8,
						'I' => 1,		'i'	=> 1,
						'J' => 20,		'j' => 20,
						'K' => 18,		'k' => 18,
						'L' => 7,		'l' => 7,
						'M' => 6,		'm' => 6,
						'N' => 13,		'n' => 13,
						'O' => 20,		'o' => 20,
						'P' => 14,		'p' => 14,
						'Q'	=> 15,		'q'	=> 15,
						'R' => 19,		'r' => 19,
						'S' => 11,		's' => 11,
						'T' => 12,		't' => 12,
						'U' => 20,		'u'	=> 20,
						'V' => 0,		'v' => 0,
						'W' => 4,		'w' => 4,
						'X' => 20,		'x' => 20,
						'Y' => 3,		'y' => 3,
						'Z' => 20,		'z' => 20
					);
					
	my @condProbIn = (
		[-1.368801, -1.500845, -1.643014, -1.621860, -2.088331, -1.503038, -1.619388, -1.552819, -2.756840, -1.505897, -1.887070, -1.996554, -1.374318, -1.791759, -2.662588, -1.563976, -1.540445, -3.020425, -1.791759, -1.329136, -9.210340],
		[-1.730443, -1.582964, -1.944339, -1.848918, -1.837016, -1.733216, -1.670682, -1.722378, -2.197225, -1.953209, -1.956063, -1.868721, -2.472930, -2.148434, -2.374906, -1.969441, -1.828127, -1.767662, -1.791759, -2.833213, -9.210340],
		[-2.403056, -2.474782, -2.124465, -2.197225, -2.424803, -2.304666, -2.130214, -2.394019, -2.197225, -1.979185, -2.397895, -1.952102, -2.167549, -2.995732, -2.151762, -1.969441, -2.639057, -1.921813, -2.197225, -2.427748, -9.210340],
		[-3.056357, -3.053816, -2.871680, -3.008155, -2.530163, -2.997813, -2.910372, -2.948330, -3.044523, -2.783558, -2.985682, -3.141686, -2.714092, -4.094345, -3.068053, -3.761200, -4.025352, -3.020425, -2.981120, -1.916923, -9.210340],
		[-4.576183, -4.095270, -4.152614, -3.583519, -3.341094, -3.978643, -4.209655, -4.229264, -4.143135, -4.518159, -3.342357, -3.478158, -4.418840, -2.484907, -2.374906, -3.068053, -4.034475, -4.034475, -4.034475, -2.833213, -9.210340],
		[-2.542411, -2.542990, -2.583997, -2.602690, -2.530163, -2.843663, -2.910372, -2.426455, -2.063693, -3.314186, -2.890372, -2.689701, -2.472930, -2.302585, -2.662588, -3.068053, -2.415914, -1.634130, -2.890372, -3.526361, -9.210340],
		[-3.526361, -3.348056, -3.277145, -3.382848, -3.628776, -3.777972, -2.823361, -3.410954, -3.449988, -3.825012, -4.189655, -5.087596, -3.166078, -4.094345, -3.761200, -2.374906, -3.332205, -2.614960, -3.453596, -3.526361, -9.210340],
		[-1.830745, -1.770706, -1.911904, -1.791759, -2.019338, -1.665008, -1.781907, -1.711567, -2.197225, -1.979185, -1.920971, -1.996554, -1.893112, -2.014903, -2.374906, -2.151762, -2.639057, -2.104134, -1.280934, -2.833213, -9.210340],
		[-5.828946, -5.039732, -4.509288, -4.682131, -4.727388, -4.096426, -4.615120, -4.991404, -2.756840, -4.007333, -4.595120, -3.295837, -4.418840, -4.094345, -4.618729, -3.761200, -4.025352, -3.020425, -4.618729, -4.618729, -9.210340],
		[-3.104366, -3.322080, -2.817612, -2.947530, -3.628776, -3.873282, -3.516508, -3.299728, -2.533697, -2.726399, -2.803360, -3.478158, -2.809403, -4.094345, -1.969441, -3.761200, -2.233592, -3.713572, -3.145092, -3.526361, -9.210340],
		[-3.814043, -3.653437, -3.564827, -3.478158, -2.781478, -3.777972, -4.209655, -3.570018, -3.449988, -3.131865, -2.197225, -3.141686, -3.166078, -3.401197, -1.969441, -3.761200, -3.332205, -3.713572, -2.890372, -3.526361, -9.210340],
		[-4.124197, -3.766766, -3.319704, -3.834833, -3.117950, -3.777972, -5.308268, -3.846272, -2.351375, -4.007333, -3.342357, -2.448539, -5.111988, -3.401197, -3.068053, -2.662588, -2.415914, -3.020425, -2.890372, -3.526361, -9.210340],
		[-3.477570, -4.346584, -3.510760, -3.382848, -4.034241, -3.536810, -3.362358, -3.718438, -3.449988, -3.314186, -3.342357, -5.087596, -3.725693, -2.302585, -3.761200, -3.761200, -4.025352, -3.713572, -3.649876, -2.140066, -9.210340],
		[-4.912655, -5.039732, -5.356586, -5.780744, -3.117950, -4.384108, -5.308268, -4.857872, -4.143135, -5.616771, -4.595120, -4.394449, -3.320228, -4.667519, -3.761200, -3.068053, -2.926739, -4.667519, -4.667519, -3.526361, -9.210340],
		[-6.116628, -5.599348, -4.845761, -5.087596, -3.341094, -5.077255, -5.308268, -5.551020, -5.000663, -3.825012, -3.496508, -4.394449, -5.111988, -4.094345, -5.000663, -3.761200, -3.332205, -5.000663, -5.000663, -5.000663, -9.210340],
		[-5.018015, -5.193882, -4.663439, -5.780744, -4.034241, -5.482720, -3.921973, -5.327876, -4.143135, -5.616771, -5.288267, -3.988984, -5.111988, -3.401197, -3.761200, -5.000663, -5.000663, -5.000663, -2.890372, -3.526361, -9.210340],
		[-4.730333, -4.788417, -5.068904, -5.780744, -4.736512, -4.566429, -4.615120, -5.551020, -4.143135, -3.825012, -4.595120, -3.478158, -5.111988, -2.995732, -3.068053, -4.736512, -4.736512, -4.736512, -4.736512, -2.833213, -9.210340],
		[-6.522093, -5.039732, -4.663439, -5.087596, -5.048291, -4.096426, -4.209655, -5.327876, -3.449988, -5.616771, -5.288267, -4.394449, -5.111988, -5.048291, -5.048291, -5.048291, -5.048291, -3.020425, -2.197225, -5.048291, -9.210340],
		[-6.116628, -5.887030, -5.762052, -5.871491, -5.871491, -6.175867, -5.871491, -5.327876, -5.871491, -5.871491, -5.288267, -5.087596, -5.871491, -5.871491, -5.871491, -3.761200, -5.871491, -3.020425, -5.871491, -3.526361, -9.210340],
		[-5.018015, -6.292495, -5.356586, -4.171306, -4.034241, -6.175867, -5.308268, -6.244167, -5.235503, -5.616771, -5.288267, -5.087596, -3.725693, -4.094345, -5.235503, -3.761200, -3.332205, -5.235503, -2.890372, -5.235503, -9.210340],
		[-9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340]
	);
	
	my @condProbOut = (
		[-2.179141, -2.382628, -2.691243, -2.373354, -2.313635, -2.151762, -2.647946, -2.049841, -2.407419, -2.743768, -2.564949, -2.797281, -2.365105, -2.960105, -2.197225, -2.437116, -3.012262, -2.649822, -2.689358, -2.270980, -9.210340],
		[-2.568606, -2.025953, -3.065937, -2.830112, -2.564949, -2.417465, -2.329493, -2.359996, -2.786909, -2.502606, -2.970414, -2.648861, -2.908721, -2.960105, -2.756840, -2.711553, -2.724580, -2.888233, -2.661959, -2.676445, -9.210340],
		[-3.533687, -3.722402, -2.691243, -3.289645, -3.124565, -3.068053, -2.647946, -3.199747, -2.643808, -2.576714, -2.564949, -2.925115, -3.314186, -3.653252, -3.449988, -3.576550, -3.705409, -3.255958, -3.633820, -4.468204, -9.210340],
		[-2.905078, -3.175858, -2.978925, -2.442347, -3.124565, -2.608521, -4.034241, -3.099663, -2.579269, -2.656757, -2.564949, -2.925115, -3.853183, -2.697741, -2.063693, -2.947942, -3.299944, -3.438279, -2.635291, -3.133203, -9.210340],
		[-4.109051, -4.174387, -4.077538, -4.388257, -3.817712, -4.859812, -4.727388, -4.198276, -4.658711, -4.043051, -3.663562, -4.406719, -5.799093, -3.876396, -3.449988, -3.864232, -4.110874, -4.200419, -4.075653, -3.775057, -9.210340],
		[-2.905078, -2.984803, -2.978925, -2.830112, -3.817712, -2.913902, -3.117950, -2.886089, -3.049273, -3.195753, -2.564949, -3.377100, -3.601868, -3.316780, -3.449988, -3.864232, -3.194583, -3.507272, -3.564827, -3.187270, -9.210340],
		[-4.226834, -3.722402, -3.384390, -5.081404, -4.510859, -3.943522, -3.341094, -4.352426, -3.742420, -3.195753, -4.356709, -4.224398, -4.546329, -4.164078, -4.143135, -4.269698, -4.110874, -3.661422, -3.787970, -3.880418, -9.210340],
		[-2.211931, -2.336108, -2.519393, -2.730029, -2.564949, -2.294863, -2.935628, -2.272985, -2.712801, -2.197225, -2.564949, -2.720320, -3.026504, -3.470931, -1.945910, -2.765620, -2.789118, -2.745132, -2.689358, -2.942148, -9.210340],
		[-3.361836, -3.555348, -2.755782, -3.001963, -3.817712, -3.250375, -3.117950, -3.505128, -3.560099, -2.944439, -3.663562, -3.377100, -3.401197, -3.876396, -3.044523, -3.576550, -3.099273, -3.661422, -3.500288, -3.775057, -9.210340],
		[-3.908380, -3.481240, -2.898882, -3.289645, -3.412247, -3.607049, -2.781478, -3.199747, -3.154634, -3.349904, -4.356709, -3.531250, -4.007333, -3.183249, -4.143135, -4.046554, -3.551258, -4.200419, -4.886583, -3.512693, -9.210340],
		[-4.514516, -4.734003, -3.672072, -3.982792, -3.817712, -3.761200, -4.727388, -4.352426, -4.658711, -5.141664, -2.277267, -4.224398, -4.412798, -4.569543, -4.143135, -5.655992, -5.497168, -6.146329, -4.663439, -4.131732, -9.210340],
		[-3.087399, -2.753002, -2.372789, -2.683509, -2.901422, -2.913902, -2.935628, -2.848349, -2.712801, -2.656757, -2.564949, -2.432638, -2.272732, -2.697741, -3.449988, -2.883403, -2.406126, -3.101807, -3.181834, -2.745438, -9.210340],
		[-2.179141, -2.536778, -2.285778, -3.135494, -3.817712, -2.662588, -2.781478, -2.678450, -2.260816, -2.656757, -2.277267, -1.796649, -1.847849, -2.043814, -2.533697, -1.992430, -2.278292, -2.176037, -2.401676, -2.052290, -9.210340],
		[-4.003690, -3.817712, -3.854394, -3.209602, -3.124565, -3.607049, -3.628776, -4.352426, -3.965564, -3.062222, -3.663562, -3.451208, -3.273364, -3.470931, -3.449988, -3.091043, -3.551258, -3.255958, -3.439664, -3.081910, -9.210340],
		[-4.360365, -4.734003, -4.770685, -3.695110, -3.817712, -4.859812, -4.727388, -3.946961, -4.253246, -5.141664, -4.356709, -5.323010, -4.882802, -4.569543, -4.595280, -4.962845, -5.497168, -4.536891, -5.579730, -4.468204, -9.210340],
		[-3.087399, -3.175858, -3.384390, -3.066501, -2.719100, -3.761200, -3.341094, -3.253814, -3.272417, -3.532226, -4.356709, -3.243568, -2.828678, -2.697741, -3.449988, -2.477938, -2.858111, -3.101807, -2.977040, -3.369592, -9.210340],
		[-3.821369, -3.347709, -3.672072, -3.577327, -3.124565, -3.250375, -3.341094, -3.436136, -2.953963, -3.195753, -4.356709, -2.925115, -3.273364, -3.316780, -4.143135, -3.016934, -3.705409, -3.843744, -2.689358, -2.745438, -9.210340],
		[-2.809768, -2.862201, -2.573460, -3.066501, -2.564949, -2.913902, -2.242481, -2.742988, -2.866951, -3.195753, -4.356709, -2.971635, -2.521948, -2.372319, -2.533697, -2.611469, -3.194583, -2.968275, -1.842060, -2.034591, -9.210340],
		[-2.722756, -2.509379, -2.824775, -2.136965, -2.313635, -2.844909, -2.242481, -2.560667, -2.579269, -3.755369, -2.747271, -2.925115, -2.621039, -2.429477, -3.449988, -2.360155, -1.913649, -1.715513, -2.444236, -2.819546, -9.210340],
		[-2.499613, -2.719100, -3.854394, -2.830112, -2.208274, -2.662588, -2.530163, -3.008691, -3.049273, -2.576714, -2.410799, -2.683953, -2.466888, -2.266958, -2.533697, -2.947942, -2.164964, -2.103278, -3.014781, -3.592736, -9.210340],
		[-9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340, -9.210340]
	);
	

### H.2.2 Probability table reader ###
	
	# A provision to read in alternate tables from another source
	
	if (defined $opts->{'p'}) {
		open (CONDPROBFILE, $probFileName) or die "Betascan dies in calcProbs: can't open alternate conditional probability file $probFileName \n";
		my @condProbFile = <CONDPROBFILE>;
		close CONDPROBFILE;
		my $mode = 0;
		my $rows = 0;
		my @columns = ();
		print MAINERR "Reading conditional probability file $probFileName \n";
		foreach my $line (@condProbFile) {
			print MAINERR "Line: $line \n";
			# discard blank lines
			if ($line =~ /^\s*$/) {
				print MAINERR "is blank\n";
				next;
			}
			# discard comment lines
			elsif ($line =~ /^\s*#/
			) 
			{
				print MAINERR "is a comment \n";
				next;
			}
			elsif ($line =~ /^\s*EXPOSEDPROBS\s*$/) {
				print MAINERR "Starting exposed probs \n";
				unless ($mode == 0) {
					print MAINERR "ERROR: we weren't done with something else!! \n";
				}
				$mode = 11;
				$rows = 0;
				next;
			}
			elsif ($line =~ /^\s*BURIEDPROBS\s*$/) {
				print MAINERR "Starting buried probs \n";
				unless ($mode == 0) {
					print MAINERR "ERROR: we weren't done with something else!! \n";
				}
				$mode = 21;
				$rows = 0;
				next;
			}
			elsif ( ($mode % 10) == 1) {
				@columns = split( '\t', $line);
				shift @columns;
				$mode ++;
			}
			elsif ( ($mode % 10) == 2) {
				my @data = split( '\t', $line);
				my $rowheader = shift @data;
				for (my $columnNum = 0; $columnNum < scalar(@data); $columnNum++) {
					if ($mode == 12) {
						# exposed
						$condProbOut[$gLookupTable{$rowheader}][$gLookupTable{$columns[$columnNum]}] = $data[$columnNum];
					}
					elsif ($mode == 22) {
						# buried
						$condProbIn[$gLookupTable{$rowheader}][$gLookupTable{$columns[$columnNum]}] = $data[$columnNum];
					}
					# if we ever do parallel tables, here's where we read them
				} # end for
				$rows ++;
				if ($rows == scalar (@columns)) {
					# we should be done now
					$mode = 0;
					print MAINERR "Done with that section\n";
				}
			} 
			else {
				print MAINERR "ERROR: Line not expected!! \n";
			}
		} # end foreach my $line
		print MAINERR "Done reading conditional probability file \n";
	} # end if defined $opts->{'p'}
	
### H.2.3 Other initializations ###
	
	my $proteinName = '';
	my @rsequence = ();
	my @bestProteinScores;
	for (my $ii = 0; $ii < 300; $ii++) {
		$bestProteinScores[$ii] = -500;
	}
	
	print MAINERR "In calcProbs okay \n";
	print STDERR "\tCalculator: ";
	
### H.3 Singleton preparatory calculations ###	
		
	my $singleInRef;
	my $singleOutRef;
	($singleInRef, $singleOutRef) = singleTables(\@condProbIn, \@condProbOut, \%gLookupTable);
	my @singleIn = @$singleInRef;
	my @singleOut = @$singleOutRef;
	
	print STDERR "ready. \n";
	
### H.4 Main calculation loop ###

	foreach my $mySequenceName (keys %$seqHashRef) {
		print STDERR "\t\t$mySequenceName";
		my $mySequence = $seqHashRef->{$mySequenceName};
		my $myAltSequence;
		if (defined $opts->{'a'}) {
			$myAltSequence = $seqHashRef->{$opts->{'a'}};
		}
		my $length = length $mySequence;
		for (my $strandLen = 2; $strandLen <= $maxBetaLength; $strandLen++) {
			for (my $first = 0; $first < 2; $first++) {
				print (".");
			
### H.4.1 Initializations ###
				my @scores;
				my $pos1;
				my $pos2;
				for ($pos1 = 0; $pos1 < 2000; $pos1++) {
					$scores[$pos1] = -500;
				}

				my @orientNames = ("in-out", "out-in"); 
				my $twoDigit = sprintf("%02d", $strandLen);
				my $commonSuffix = $orientNames[$first] . " " . $twoDigit . " residue.txt";
### H.4.2 Open output files ###
				# the singleton chart
				my $singletonFileName =  $mySequenceName . " prion single strand " . $commonSuffix;			
				open (OUT3, ">$singletonFileName") or die "Betascan dies in calcProbs: can't open singleton $singletonFileName \n";
			
				# the pairwise charts, parallel
				my $combinedFileName = $mySequenceName . " prion combined scores " . $commonSuffix;		
				open (OUT4, ">$combinedFileName") or die "Betascan dies in calcProbs: can't open combined $combinedFileName \n";
				my $forwardFileName =  $mySequenceName . " prion forward scores " . $commonSuffix;			
				open (OUT5, ">$forwardFileName") or die "Betascan dies in calcProbs: can't open forward $forwardFileName \n";
			
				# the pairwise charts, antiparallel
				if (defined $opts->{'A'}) {
					my $antiCombinedFileName =  $mySequenceName . " antiparallel prion combined scores " . $commonSuffix;		
					open (OUT10, ">$antiCombinedFileName") or die "Betascan dies in calcProbs: can't open combined $combinedFileName \n";
					my $antiForwardFileName =  $mySequenceName . " antiparallel prion forward scores " . $commonSuffix;			
					open (OUT11, ">$antiForwardFileName") or die "Betascan dies in calcProbs: can't open forward $forwardFileName \n";
				}
### H.4.3 Singletons ###

				## start looping for the first strand
				for ($pos1 = 0; $pos1 < ($length); $pos1++) {

### H.4.3.1 Initializations ###
				
					my $initialScore = 0;
					my $antiInitialScore = 0;
					my $correctedInitialScore = 0;
					my $direction = $first;
					my $normscore;
					my $normflipscore;
					my $correction;
					my $firstDenomCorrection;
					my $antiFirstDenomCorrection;
### H.4.3.2 Singleton calculations ###

					# singleIn, singleOut, condProbIn, condProbOut are all in the scrambled format 
					# where column 0 = V, etc.
					# as delineated in gLookupTable.
					# so to get the singleIn score of valine
					# we must know the column number for valine, which is gLookupTable{'V'}
				
					if ($pos1 < ($length - $strandLen)) {
						# sum the score "initialscore", which is the singleton score of the first strand
						for (my $i = 0; $i < $strandLen; $i++) {
							my $currentResidue = substr($mySequence, $pos1 + $i, 1);
							if ($direction == 0) { $initialScore += $singleIn[$gLookupTable{$currentResidue}];  }
							else				 { $initialScore += $singleOut[$gLookupTable{$currentResidue}]; }
							$direction ^= 1;
						}								
						
						$antiInitialScore = $initialScore;
						my $strandSeq = substr($mySequence, $pos1, $strandLen);
						$correction = corrector("none", $strandSeq, 0);
						$firstDenomCorrection = corrector($model, $strandSeq, 0);
						$initialScore += $correction;
#						print "At $pos1, $firstDenomCorrection \n";
						$correctedInitialScore = $initialScore - $firstDenomCorrection;
						
						my $antiFirstCorrection = corrector("none", $strandSeq, 1);
						$antiInitialScore += $antiFirstCorrection;
						$antiFirstDenomCorrection = corrector($model, $strandSeq, 1);
						
					} # end if $pos1 < ...
					

					
					print OUT3 $pos1 . "\t" . $correctedInitialScore . "\n";
				
### H.4.4 Pairwises ###					
					for ($pos2 = 0; $pos2 < ($length); $pos2++) {

### H.4.4.1 Distance exclusion and initialization ###				
						## distance exclusion parameters: minloop and maxloop
											
						if (($pos1 -$pos2) < $minLoop && ($pos2 - $pos1) < $minLoop)  {		next;		}
						if (($pos1 - $pos2) > $maxLoop || ($pos2 - $pos1) > $maxLoop) {		next;		}
						my $score = 0;
						my $finalScore = 0;
						my $antiScore = 0;
						my $antiFinalScore = 0;
						$direction = $first;

### H.4.4.2 Pairwise calculations ###						
						## sum the pairwise scores
						if ($pos1 < ($length - $strandLen) && $pos2 < ($length - $strandLen)) {
							## sum the score "score", which is the pairwise score of the two strands
							## note: pairwise only
							for (my $nn = 0; $nn < $strandLen; $nn++) {
								my $firstResidue = substr($mySequence, $pos1 + $nn, 1);
								my $secondResidue = substr($mySequence, $pos2 + $nn, 1);
								if (defined $opts->{'a'}) {
									$secondResidue = substr($myAltSequence, $pos2 + $nn, 1);
								}
								if ($direction == 0) { $score += $condProbIn[$gLookupTable{$secondResidue}][$gLookupTable{$firstResidue}];  }
								else				 { $score += $condProbOut[$gLookupTable{$secondResidue}][$gLookupTable{$firstResidue}]; }
								
								## Bonuses and penalties added here
								if (defined $opts->{'y'}) {
									$score += pairBonuses($firstResidue, $secondResidue);
								}
								## Line below is the one that was missing
								$direction ^= 1;
							}
							## normalize the pairwise score by length and orientation
							# $score -= $normscore;
							# $score -= $normflipscore;
							my $secondStrandSeq = substr($mySequence, $pos2, $strandLen);
							if (defined $opts->{'a'}) {
								$secondStrandSeq = substr($myAltSequence, $pos2, $strandLen);
							}
							my $secondCorrection = corrector("none", $secondStrandSeq, 0);
							$score += $secondCorrection;
						
							## for the joint prob, normalize by the first strand's prob too.
							$finalScore = $score + $initialScore;
							
							## for the CORRECTED joint prob, we must calculate log-odds.
							
							my $secondDenomCorrection = corrector($model, $secondStrandSeq, 0);
							
							$score -=		$secondDenomCorrection;
							$finalScore -=	$secondDenomCorrection;
							$finalScore -=  $firstDenomCorrection;

							if (defined $opts->{'A'}) {
								## if we do the antiparallel calculations... let's do them
								for (my $pp = 0; $pp < $strandLen; $pp++) {
									my $firstResidue = substr($mySequence, $pos1 + $pp, 1);
									my $secondResidue = substr($mySequence, 
																($pos2 + ($strandLen - 1)) - $pp, 1);
									if (defined $opts->{'a'}) {
										$secondResidue = substr($myAltSequence, 
																($pos2 + ($strandLen - 1)) - $pp, 1);
									}
									if ($direction == 0) {
										$antiScore += $condProbIn[$gLookupTable{$secondResidue}][$gLookupTable{$firstResidue}];
									}
									else {
										$antiScore += $condProbOut[$gLookupTable{$secondResidue}][$gLookupTable{$firstResidue}];
									}
									$direction ^= 1;
								}
								my $antiSecondCorrection = corrector("none", $secondStrandSeq, 1);
								$antiScore += $antiSecondCorrection;
								$antiFinalScore = $antiScore + $antiInitialScore;
								
								my $antiSecondDenomCorrection = corrector($model, $secondStrandSeq, 1);
								
								$antiScore		-= $antiSecondDenomCorrection;
								$antiFinalScore -= $antiSecondDenomCorrection;
								$antiFinalScore -= $antiFirstDenomCorrection;
							}
							## calculations complete.
						} # end if $pos1 < ...
					
### H.4.5 Outputs ###	
					
						## "combined", throughout, means add the singleton score to the pairwise score.
						## since they are log scores, that's a multiplication, e.g.,
						## Prob (strand forming) * Prob (strand pairing)
						## Note that this assumes independence of strand formation and strand pairing
						## which may be a bad assumption when dealing with edge effects.
					
						## Forward and forward combined scores
						print OUT5 $pos1 ."\t". $pos2 ."\t". $score ."\n";
						print OUT4 $pos1 ."\t". $pos2 ."\t". $finalScore ."\n";
						
						if (defined $opts->{'A'}) {
							print OUT10 $pos1 ."\t". $pos2 ."\t" . $antiScore . "\n";
							print OUT11 $pos1 ."\t". $pos2 ."\t" . $antiFinalScore . "\n";
						}
						
					} # end for my $pos2...
				} # end for my $pos1...
			
### H.4.6 Completions ###
			
				## let's close it out.
				close OUT3;
				close OUT4;
				close OUT5;
				if (defined $opts->{'A'}) {
					close OUT10;
					close OUT11;
				} # end if defined $opts->{'A'}				
			} # end for my $first...
		} # end for my $strandLen...
		print STDERR "\n";
	} # end foreach my $mySequenceName (keys %proteinSequenceHash)

} # end sub calcProbs

##################################################################
### I. sub readSequence                                        ###
### reads in sequences from FASTA file                         ###
### returns refs to two hashes                                 ###
### %proteinSeqHash contains all the sequences                 ###
### %proteinSeqNameHash contains the full names of the seqs    ###
### the first word is used as the ID and must be unique        ###
##################################################################

sub readSequence {
	my $proteinSequenceFileName = shift;
	
	# If I want the sequences accessible outside this subroutine
	# I'll have to move these declarations to global scope...
	# which I sort of did, as I now pass these by reference.
	my %proteinSeqHash;
	my %proteinSeqFullName;
	my $readInSeqName = '';
	my $readInSeq = '';
	
	# Open protein sequence file
	open (PROTEINSEQFILE, $proteinSequenceFileName) or die "Betascan dies in readSequences: Error opening main sequence file.\n";
	my @proteinSeqFile = <PROTEINSEQFILE>;
	close PROTEINSEQFILE;
	
	print MAINERR "In readSequence, drained file okay \n";
	print STDERR "\tSequences to be processed:\n";
	# Process FASTA file
	foreach my $line (@proteinSeqFile) {
		print MAINERR "\tLine: $line";
		# discard blank lines
		if ($line =~ /^\s*$/) {
			print MAINERR "is blank\n";
			next;
		}
		# discard comment lines
		elsif ($line =~ /^\s*#/
		) 
		{
			print MAINERR "is a comment \n";
			next;
		}
		# fasta header line: first word becomes key of hash
		elsif ($line =~ /^>/) {
			# dump old sequence, if not already done
			if ($readInSeqName ne '') {
				# get rid of whitespace in case we have any
				$readInSeq =~ s/\s//g;
				# check it
				print MAINERR "\t\tSequence: $readInSeq \n";
				# send to hash
				$proteinSeqHash{$readInSeqName} = $readInSeq;
			}
			# new sequence
			$readInSeq = '';
			# if the routine below fails to get the right sequence name we have a problem, hence the dummy setting
			$readInSeqName = 'PROBLEM';
			my $fullName = substr($line, 1);
			my @splitLine = split(/ /,$line);
			my $shortName = substr($splitLine[0], 1);
			chomp $shortName;
			# use shortName as hash key
			# if already exists exit, as rest of code requires short names at present
			# store fullName as well
			print STDERR "\t\t". $shortName;
			print MAINERR "\t\t" . $shortName;
			$readInSeqName = $shortName;
			$proteinSeqFullName{$shortName} = $fullName;
			if (exists($proteinSeqHash{$readInSeqName})) {
				die "Betascan dies in readSequences: Two sequence titles in the FASTA file have the identical first word ".$shortName.". Please fix that.\n";
			}
			print STDERR "\n";
			print MAINERR "\n";
		} # end elsif ($line =~ /^>/) 
		else {
			print MAINERR " is a sequence line\n";
			$readInSeq .= $line;
			# in case that line is old style last line where * = STOP
			if ($line =~ /^.*\*$/) {
				chop $readInSeq;
			}
		}
	} # end foreach my $line
	# put the last sequence in
	# get rid of whitespace in case we have any
	$readInSeq =~ s/\s//g;
	# check it
	print MAINERR "Last sequence: $readInSeq \n";
	# send to hash
	$proteinSeqHash{$readInSeqName} = $readInSeq;
	
	if (defined $opts->{'a'}) {
		unless (defined $proteinSeqHash{$opts->{'a'}}) {
			die "Betascan dies in readSequences: -a (two protein runs) selected but named sequence not found in sequence file.\n";
		}
	}
	
	return (\%proteinSeqHash, \%proteinSeqFullName);
} # end sub readSequences

##################################################################
### J. sub singleTables                                        ###
### takes the conditional probability tables from BETAWRAP     ###
### calculates the singleton tables from them                  ###
### prints them to files and returns them as well              ###
##################################################################

sub singleTables {
	my $condProbInRef = shift;
	my $condProbOutRef = shift;
	my $gLookupTableRef = shift;
	my @condProbInS = @$condProbInRef;
	my @condProbOutS = @$condProbOutRef;
	my @realResidues = ('A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y');
	
	## we calculate singleton scores from the pairwise scores.
	## due to fungooed statistics, we exclude the X (bad) residues from this calculation.
	my $totalIn = 1;
	my $totalOut = 1;
	## the choice of calculating these on the zeroth row and column is arbitrary.
	## any row/column pair would work as long as they are the ith row and ith column.
	## the same results would be obtained.
	## however, because the numbers for X are all the same, their value would be inflated,
	## hence their exclusion.
	
	my @singleIn = ( 1 );
	my @singleOut = ( 1 );
	for (my $jj = 1; $jj < 20; $jj ++) {
		my $f = exp($condProbInS[0][$jj]);
		my $f2 = exp($condProbInS[$jj][0]);
		$singleIn[$jj] = $f2 / $f;
		$totalIn += $f2 / $f;
		
		$f = exp($condProbOutS[0][$jj]);
		$f2 = exp($condProbOutS[$jj][0]);
		$singleOut[$jj] = $f2 / $f;
		$totalOut += $f2 / $f;
	}
	
	## normalize and put back in log form.
	
	for (my $kk = 0; $kk < 20; $kk++) {
		$singleIn[$kk] /= $totalIn;
		$singleOut[$kk] /= $totalOut;
		$singleIn[$kk] = log($singleIn[$kk]);
		$singleOut[$kk] = log($singleOut[$kk]);
	}
	
	## output to files.
	
	my $singleInFile = "singleInFreqs-" . $idTime . ".txt";
	my $singleOutFile = "singleOutFreqs-" . $idTime . ".txt";
	open(SINGLEINFREQ, ">$singleInFile") or die ("Betascan dies in calcProbs: can't open singleInFreqs.txt\n");
	foreach my $freqInResidue (@realResidues) {
		print SINGLEINFREQ $freqInResidue . "\t" . $singleIn[$gLookupTableRef->{$freqInResidue}] . "\n";
	}
	close SINGLEINFREQ;
	
	open(SINGLEOUTFREQ, ">$singleOutFile") or die ("Betascan dies in calcProbs: can't open singleOutFreqs.txt\n");
	foreach my $freqOutResidue (@realResidues) {
		print SINGLEOUTFREQ $freqOutResidue . "\t" . $singleOut[$gLookupTableRef->{$freqOutResidue}] . "\n";
	}
	close SINGLEOUTFREQ;

	return(\@singleIn, \@singleOut);
}

##################################################################
### K. sub makeHTML                                            ###
### Be able to find all the generated files on one page        ###
##################################################################

sub makeHTML {

### K.1 Read in large number of arguments ###
	my $startDir = shift;
	my $idTime = shift;
	my $fileName = shift;
	my $timeString = shift;
	my $optString = shift;
	my $inputFile = shift;
	my $minLoop = shift;
	my $maxLoop = shift;
	my $imageOpt = shift;
	my $model = shift;
	my $weightFileName = shift;
	my $opts = shift;
	my $betaLength = shift;
	my $strLength = shift;
	my $proteinNamesRef = shift;

### K.2 Open file and report all settings ###
	
	open (WEBFILE, ">$fileName") or die "Betascan dies in makeHTML: can't open webpage $fileName!\n";
	
	# Header
	print WEBFILE "<html>\n<head>\n<title>\nBetascan result page\n</title>\n</head>\n\n<body>\n";
	print WEBFILE "<h1>Betascan results<br /></h1>\n";
	print WEBFILE "<h4>Calculation run at " . $timeString . "<br />\n";
	print WEBFILE "Summary of options: " . $optString . "</h4><br />\n";
	
	# Mandatory settings
	print WEBFILE "     FASTA sequence file:          " . $opts->{'n'}. "<br /><br />\n";
	print WEBFILE "     Maximum beta strand length:   " . $betaLength . " residues<br />\n";
	if ($strLength > 0) {
		print WEBFILE "     Length of strand evaluated:   " . $strLength . " residues<br />\n";
	}
	else {
		print WEBFILE "     Length of strand evaluated:   full length of each sequence<br />\n";
	}
	
	# Optional switches
	
	# -m
	print WEBFILE "		Minimum loop length allowed:  " . $minLoop     . " residues";
	if ($minLoop == 0) {
		print WEBFILE " (default), complete interstrand pairing allowed";
	}
	print WEBFILE "<br />\n";
	
	# -M
	print WEBFILE "     Maximum loop length allowed:  " . $maxLoop     . " residues";
	if ($maxLoop == 35) {
		print WEBFILE " (default)";
	}
	print WEBFILE "<br /><br />\n";
	
	# -p
	if (defined $opts->{'p'}) {
		print WEBFILE "     Probability table file:       " . $probFileName . "<br />\n";
	}
	
	# -N
	print WEBFILE "     Poisson length normalization: " . $normParam;
	if ($normParam == 3.15) {
		print WEBFILE " (default)";
	}
	print WEBFILE "<br />\n";
	
	# -L
	if (defined $opts->{'L'}) {
		print WEBFILE "     Length normalization deactivated <br />\n";
	}
	
	# -o
	if (defined $opts->{'o'}) {
		print WEBFILE "Normalization by orientation used <br />\n";
	}
	
	# -c
	print WEBFILE "     Amino acid propensity model:  " . $model;
	if ($model eq "allproteins") {
		print WEBFILE " (default)";
	}
	print WEBFILE "<br />\n";
	
	# -w
	print WEBFILE "     Amino acid propensity file:   " . $weightFileName;
	if ($weightFileName eq "BSweights.txt") {
		print WEBFILE " (default)";
	}
	print WEBFILE "<br />\n";
	
	# -y
	if (defined $opts->{'y'}) {
		print WEBFILE "     Using stacking bonuses <br />\n";
	}
	
	# -A 
	if (defined $opts->{'A'}) {
		print WEBFILE "     Including antiparallel pairs <br />\n";
	}
	
	print WEBFILE "<br />\n";
	
	# -i
	print WEBFILE "     Image format:                 " . $imageOpt    . "<br />\n";
	
	# -s12gdf
	if (defined $opts->{'s'}) {
		print WEBFILE "All pre-maxima graphics calculations supressed <br />\n";
	}
	if (defined $opts->{'1'}) {
		print WEBFILE "Pairwise full-set calculations and graphics generation supressed <br />\n";
	}
	if (defined $opts->{'2'}) {
		print WEBFILE "Singleton full-set calculations and graphics generation supressed <br />\n";
	}
	if (defined $opts->{'g'}) {
		print WEBFILE "All full-set graphics generation supressed <br />\n";
	}
	if (!defined $opts->{'F'}) {
		print WEBFILE "Generating only most-used graphs <br />\n";
	}
	unless (defined $opts->{'f'}) {
		print WEBFILE "Maxima calculated and graphed <br />\n";
	}
	unless (defined $opts->{'d'}) {
		print WEBFILE "Data sets autosorted <br />\n";
	}

	print WEBFILE "<br /><br /><br />\n";
	
### K.3 Results for each protein ###

	foreach my $protein (@$proteinNamesRef) {
	
### K.3.1 Initializations ###

		my $timeStamp = sprintf("%05d", $idTime);
		my $dirTag;	
		if ($address ne 'DEADBEEF') {
			$dirTag =  $address . $protein . "-". $timeStamp;
		}
		else {
			$dirTag =  $protein . "-" . $timeStamp;
		}
		my $mainDir = $cwd . "/" . $dirTag;

### K.3.2 Header ###
		
		print WEBFILE "<hr>\n";
		print WEBFILE "<h2>" . $protein ."<br /></h2>\n";
		
### K.3.3 Singletons ###		
		print WEBFILE "<h4>Singletons: </h4>\n";
		# Raw set always generated
		unless (defined $opts->{'d'}) {
			print WEBFILE "Data files <a href=\"$mainDir/txt/single/\">(dir)</a>, ";
		}
		print WEBFILE "compiled data set ";
		webLink($mainDir, $startDir, "$protein prion single strand complete swap.txt");
		print WEBFILE "<br />\n";
		# Log-odds graphs
		unless (defined $opts->{'s'}) {			
			unless (defined $opts->{'2'} || defined $opts->{'g'}) {
				print WEBFILE "Log-odds graph: ";
				webLink($mainDir, $startDir, "$protein prion single strand complete swap.txt.pdf");
				print WEBFILE "<br />\n";
			}
		}
		# Maxima graphs
		unless (defined $opts->{'f'}) {
			print WEBFILE "<b>Maxima-- </b> ";
			webLink ($mainDir, $startDir, "$protein prion single strand complete maxima.txt");
			webLink ($mainDir, $startDir, "$protein prion single strand complete maxima.txt.pdf");
		}
		print WEBFILE "<br />\n";
		
### K.3.4 Pairwise ###
		print WEBFILE "<h4>Pairwise: </h4>\n";
		# Raw set always generated
		unless (defined $opts->{'d'}) {
			print WEBFILE "Data files - combined ";
			print WEBFILE "<a href=\"$mainDir/txt/combined/in-out/\">(in-out dir)</a>";
			print WEBFILE "<a href=\"$mainDir/txt/combined/out-in/\">(out-in dir)</a>";
			print WEBFILE " - forward ";
			print WEBFILE "<a href=\"$mainDir/txt/forward/in-out/\">(in-out dir)</a>";
			print WEBFILE "<a href=\"$mainDir/txt/forward/out-in/\">(out-in dir)</a>";
			print WEBFILE "<br />\n";
		}
		# Log-odds graphs 
		unless (defined $opts->{'s'}) {
			unless (defined $opts->{'1'} || defined $opts->{'g'} || !defined $opts->{'F'}) {
				print WEBFILE "<h4>          Parallel: </h4>\n";
				my @shortbases = ("combined", "forward");
				my @orients = ("in-out", "out-in");
				my @images = ("pdf","jpg","tiff");
				foreach my $shortbase (@shortbases) {
					print WEBFILE "Log-odds graphs - " . $shortbase . "<br />\n";
					foreach my $orient (@orients) {
						print WEBFILE "        " . $orient . "<br />\n";
						foreach my $image (@images) {
							print WEBFILE "               " . $image . " ";
							for (my $loop = 2; $loop <= $maxBetaLength; $loop++) {
								my $twoDigit = sprintf("%02d", $loop);
								webLink ($mainDir, $startDir, "$protein 3dranged prion $shortbase scores $orient $twoDigit residue.txt.$image");
							}
							print WEBFILE "<br />\n";
						} # end foreach my $image...
					} # end foreach my $orient...
				} # end foreach my $shortbase...
				if (defined $opts->{'A'}) {
					print WEBFILE "<h4>          Antiparallel: </h4>\n";
					foreach my $shortbase (@shortbases) {
						print WEBFILE "Log-odds graphs - " . $shortbase . "<br />\n";
						foreach my $orient (@orients) {
							print WEBFILE "        " . $orient . "<br />\n";
							foreach my $image (@images) {
								print WEBFILE "               " . $image . " ";
								for (my $loop = 2; $loop <= $opts->{'b'}; $loop++) {
									my $twoDigit = sprintf("%02d", $loop);
									webLink ($mainDir, $startDir, "$protein 3dranged antiparallel prion $shortbase scores $orient $twoDigit residue.txt.$image");
								}
								print WEBFILE "<br />\n";
							} # end foreach my $image...
						} # end foreach my $orient...
					} # end foreach my $shortbase...
				} # end if defined opts->A
			} # end unless defined opts->1
		} # end unless defined opts->s
		
		# Maxima graphs
		unless (defined $opts->{'f'}) {
			print WEBFILE "<b>Maxima--</b> <i>Parallel:</i>";
			webLink ($mainDir, $startDir, "$protein 3dranged prion combined scores complete maxima.txt");
			webLink ($mainDir, $startDir, "$protein 3dranged prion combined scores complete maxima.txt.pdf");
			if (defined $opts->{'A'}) {
				print WEBFILE ", <i>Antiparallel:</i>";
				webLink ($mainDir, $startDir, "$protein 3dranged antiparallel prion combined scores complete maxima.txt");
				webLink ($mainDir, $startDir, "$protein 3dranged antiparallel prion combined scores complete maxima.txt.pdf");
			}
		}
		
		print WEBFILE "<br /><br />\n";
### K.3.5 Error files ###
		unless (defined $opts->{'d'}) {
			print WEBFILE "Matlab diary files: <a href=\"$mainDir/diary/\">(dir)</a><br />\n";
		}
	} # end foreach my $protein
	print WEBFILE "Matlab log file: <a href=\"$startDir/matlab-$idTime.log\">(txt)</a><br />\n";
	print WEBFILE "General log file: <a href=\"$startDir/Betascan-errors-$idTime.txt\">(txt)</a><br />\n ";

### K.4 Finish off ###
	
	print WEBFILE "<hr> \n";
	my $finish = time;
	my $finishTime = localtime($finish);
	print WEBFILE "Page generated at $finishTime<br />\n";
	my $computeTime = $finish - $runTime;
	print WEBFILE "Time to complete: $computeTime seconds<br />\n";
	close WEBFILE;
}	

##################################################################
### L. sub webLink                                             ###
### print out sorted or unsorted location as indicated         ###
### also check to see if file was generated                    ###
##################################################################

sub webLink {
		my $mainDir = shift;
		my $startDir = shift;
		my $fileName = shift;
		
#		print MAINERR "\tWebLink for " . $fileName . ": " ;
		my @hi = split(/\./, $fileName);
#		print MAINERR $hi[0] . " and " . $hi[1] . "\n";
		my $hiSize = scalar @hi - 1;
		my $suffix = $hi[$hiSize];
		
		# txt and diary are text formats; others are image formats
		if ($fileName =~ /.txt$/ || $fileName =~ /.diary$/ || $imageOpt eq "all" || $imageOpt eq $suffix) {
			print WEBFILE "<a href=\"";
			unless (defined $opts->{'d'}) {
			
				# the top directory isn't optional if we sorted
				print WEBFILE "$mainDir/$suffix/";
				
				# mid-directories
				if ($fileName =~ /single/) {
					print WEBFILE "single/";
				}	
				elsif ($fileName =~ /backward/) {
					print WEBFILE "accum-backward/";
				}
				elsif ($fileName =~ /accum/) {
					print WEBFILE "accum-forward/";
				}
				elsif ($fileName =~ /combined/) {
					print WEBFILE "combined/";
				}
				elsif ($fileName =~ /forward/) {
					print WEBFILE "forward/";
				}
				
				# low directories
				if ($fileName =~ /in-out/) {
					print WEBFILE "in-out/";
				}
				elsif ($fileName =~ /out-in/) {
					print WEBFILE "out-in/";
				}
			} 
			else {
				# we never sorted, everything's in the top directory
				print WEBFILE "$startDir/";
			}
		} #end that long if
		print WEBFILE "$fileName\">";
		unless ($fileName =~ /3dranged/ && !($fileName =~ /maxima/)) {
			print WEBFILE "($suffix)</a>";
		}
		else {
			$fileName =~ /3dranged.*(\d\d)/;
			my $digits = $1;
			print WEBFILE "($digits)</a>";
		}
}

##################################################################
### M. sub pairBonuses                                         ###
### add bonuses and penalties to a pairwise score              ###
### positive is a bonus, negative is a penalty                 ###
### returns bonus or penalty for a single pair                 ###
### there are 20x20 possible bonuses and penalties             ###
##################################################################

sub pairBonuses {
	my $firstResidue = shift;
	my $secondResidue = shift;
	my $bonus = 0;
	
	# consider coding a 20x20 array input from file
	# can be read by both here and by stitcher (if two separate programs)
	
	if ($firstResidue eq 'Y' && $secondResidue eq 'Y') {
		$bonus++;
	}
	elsif ($firstResidue eq 'Q' && $secondResidue eq 'Q') {
		$bonus++;
	}
	elsif ($firstResidue eq 'N' && $secondResidue eq 'N') {
		$bonus++;
	}

	return $bonus;
}

##################################################################
### N. sub findAttractor                                       ###
### subroutine to find the ultimate attractor                  ###
##################################################################

sub findAttractor {
	my $nextpoint = shift;
	my $nextLattice = shift;
	my $reverseZoneLattice = shift;
	
	my $currentpoint = $nextpoint;
	my $currentMax = $nextLattice->{$nextpoint};
	
	# loop
	while ($currentpoint ne $currentMax) {
		# if we can skip to the end using dynamic programming, let's do so.
		if (exists $reverseZoneLattice->{$currentMax}) {
			$currentpoint = $reverseZoneLattice->{$currentMax};
		}
		else {
			# nope, do it the hard way.
			$currentpoint = $currentMax;
		}
		$currentMax = $nextLattice->{$currentpoint};
	}
	
	return $currentpoint;
}

##################################################################
### O. sub generateMatlab                                      ###
### subroutine to generate Matlab image generation code        ### 
##################################################################

sub generateMatlab {
	# grab passed frobs
	my $idtime = shift;
	my $imgTag = shift;
	
	# default set
	if ($imgTag eq "all") {
		$imgTag = "pdf";
	}
	print MAINERR $imgTag . "\n";
	
	######
	# define all elements of the betascan plots
	######
	
	##### Headers for the four matlab files #####
	my $bsplotsingleHeader = <<BSPLOTSINGLEHEADER;
function bsplotsingle3(filename, maxBetaLength, helloThere, optString, imgTag, shift)
% bsplotsingle accepts a filename to read (filename)
% we assume the end is .txt
% we assume bsplotsingle is a singleton file
% generated by previous perl script elements of BETASCAN
% we generate a singleton plot
% and output it to filename.eps and filename.jpg
% modifying to accept variable scales 6/28/06
% mod again to print right 7/5/06
% mod to allow min and max scaling 7/11/06
% mod to allow count shifting 4/4/07
% helloThere is a flag
% 1: all values are less than 0
% 0: all values are greater than 0
% -1: values are both pos and neg

inputfile = [filename];
BSPLOTSINGLEHEADER
	my $bsplotpairHeader = <<BSPLOTPAIRHEADER;
function bsplotpair(filename, optString, shift)
% bsplotpair accepts a filename to read (filename)
% we assume the end is .txt
% we assume bsplotpair is a pairwise file
% generated by previous perl script elements of BETASCAN
% we generate a pairwise plot
% and output it to filename.eps and filename.jpg
BSPLOTPAIRHEADER
	my $bsbetasingleHeader = <<BSBETASINGLEHEADER;
function bsbetasingle2(filename, optString, shift)
% bsbetasingle takes the filename of singleton maxima
% in format of tab delimited coordinates
% position length orientation score
% with statistics to help us out
BSBETASINGLEHEADER
	my $bsbetapairHeader = <<BSBETAPAIRHEADER;
function bsbetapair2(filename, optString, shift)
% bsbetapair takes the filename of pairwise maxima
% in format of tab delimited coordinates
% position length orientation position length orientation score
% with statistics to help us out
BSBETAPAIRHEADER
	#### Common elements of the front: load data, start diary #####
	my $diaryChunk = <<DIARYCHUNK;
% whats new, pussycat?
diaryfile = [filename '.diary'];
diary(diaryfile)
diary on
filename
DIARYCHUNK
	my $dataSnarfChunk = <<DATASNARFCHUNK;
% snarf data
maindata = importdata(filename);
DATASNARFCHUNK
	#### Bodies: Unique elements of each matlab file, unchanged by frobbable matters ####
	my $bsplotsingleCalc = <<BSPLOTSINGLECALC;
% the first set of data starts at 2 and ends at maxBetaLength
% the second set of data starts at column1 and ends at column2;
column1 = maxBetaLength+1;
column2 = 2*(maxBetaLength)-1;

% now extract data from structure maindata
indata = (maindata.data(:,2:maxBetaLength));
outdata = (maindata.data(:,column1:column2));
insize = size(indata);
outsize = size(outdata);
if (abs(helloThere))
	minIn = min(min(indata));
	minOut = min(min(outdata));
	minValue = min([minIn minOut]);
end
if (helloThere <= 0)
	maxIn = max(max(indata));
	maxOut = max(max(outdata));
	maxValue = max([maxIn maxOut]);
end

% calculate shifts
shiftstart = 10 + shift
shiftend = shift + insize(1)
BSPLOTSINGLECALC
	my $bsplotsingleHello = <<BSPLOTSINGLEHELLO;
if (helloThere == 2)
    caxis([-12.5 4])
elseif (helloThere == 1)
	caxis([minValue 0])
elseif (helloThere == 0)
	caxis([0 maxValue])
else
	caxis([minValue maxValue])
end		
% shading('interp')
BSPLOTSINGLEHELLO
	my $bsplotsingleUpperBox = <<BSPLOTSINGLEUPPERBOX;
% upper box formatting
subplot('position',[0.08 0.57 0.84 0.36])
pcolor(indata')
BSPLOTSINGLEUPPERBOX
	my $bsplotsingleUpperBoxLabel = <<BSPLOTSINGLEUPPERBOXLABEL;
set(gca,'XTick',10:10:insize(1))
set(gca,'YTick',2:2:insize(2))
set(gca,'YTickLabel',2:2:(insize(2)+1))
set(gca,'XTickLabel',shiftstart:10:shiftend)
set(gca,'LineWidth',0.5)
xlabel('N-terminal residue')
ylabel('strand length')
title('single strand probabilities, odd residues buried')
colorbar('NorthOutside')
BSPLOTSINGLEUPPERBOXLABEL
	my $bsplotsingleLowerBox = <<BSPLOTSINGLELOWERBOX;
% lower box formatting
subplot('position',[0.08 0.1 0.84 0.36])
pcolor(outdata')
BSPLOTSINGLELOWERBOX
	my $bsplotsingleLowerBoxLabel = <<BSPLOTSINGLELOWERBOXLABEL;
set(gca,'XTick',10:10:outsize(1))
set(gca,'YTick',2:2:outsize(2))
set(gca,'YTickLabel',2:2:(outsize(2)+1))
set(gca,'XTickLabel',shiftstart:10:shiftend)
set(gca,'LineWidth',0.5)
xlabel('N-terminal residue')
ylabel('strand length')
title('single strand probabilities, odd residues exposed')
colorbar('NorthOutside')
BSPLOTSINGLELOWERBOXLABEL
	my $bsplotsingleText = <<BSPLOTSINGLETEXT;
% tag image with file name and timestamp
textbox1 = annotation('textbox',[0 0.95 0.2 0.05])
set(textbox1,'String',inputfile)
set(textbox1,'FitHeightToText','on')
set(textbox1,'EdgeColor','none')

textbox2 = annotation('textbox',[0.8 0.95 0.2 0.05])
set(textbox2, 'String', datestr(now))
set(textbox2, 'FitHeightToText','on')
set(textbox2, 'EdgeColor','none')
set(textbox2, 'HorizontalAlignment','right')

textbox3 = annotation('textbox',[0 0 0.2 0.05])
set(textbox3,'String',optString)
set(textbox3,'FitHeightToText','on')
set(textbox3,'EdgeColor','none')
BSPLOTSINGLETEXT
	my $bsplotpairCalc = <<BSPLOTPAIRCALC;
% now extract data
ix = [2:length(maindata)]
indata = maindata(ix, ix);
insize = length(indata)
maxValue = max(max(indata));
minValue = min(min(indata));
shiftstart = 10 + shift
shiftend = insize + shift
BSPLOTPAIRCALC
	my $bsplotpairBox = <<BSPLOTPAIRBOX;
% box formatting
pcolor(indata)
caxis([-25,10])
shading('interp')
BSPLOTPAIRBOX
	my $bsplotpairLabel = <<BSPLOTPAIRLABEL;
set(gca,'XTick',10:10:insize)
set(gca,'YTick',10:10:insize)
set(gca,'XTickLabel', shiftstart:10:shiftend)
set(gca,'YTickLabel', shiftstart:10:shiftend)
xlabel('residue number, upper strand')
ylabel('residue number, lower strand')
title('pairwise probabilities')
colorbar('EastOutside')
BSPLOTPAIRLABEL
	my $bsplotpairText = <<BSPLOTPAIRTEXT;
% tag image with file name and timestamp
textbox1 = annotation('textbox',[0 0.95 0.2 0.05])
set(textbox1,'String',filename)
set(textbox1,'FitHeightToText','on')
set(textbox1,'EdgeColor','none')

textbox2 = annotation('textbox',[0.8 0 0.2 0.05])
set(textbox2, 'String', datestr(now))
set(textbox2, 'FitHeightToText','on')
set(textbox2, 'EdgeColor','none')
set(textbox2, 'HorizontalAlignment','right')

textbox3 = annotation('textbox',[0 0 0.2 0.05])
set(textbox3,'String',optString)
set(textbox3,'FitHeightToText','on')
set(textbox3,'EdgeColor','none')
BSPLOTPAIRTEXT
	my $bsbetasingleCalc = <<BSBETASINGLECALC;
% now extract data
sourcefile = maindata.textdata;
passedData = maindata.data(1,:)
maxBeta = passedData(1);
maxSeqLength = passedData(2);
bigsizeVec = size(maindata.data);
bigsizeSize = bigsizeVec(1)
inOutData = maindata.data(2:bigsizeSize-1,:);
sizeVec = size(inOutData);
numLines = sizeVec(1)
inLines = maindata.data(bigsizeSize,1)
inPosLines = maindata.data(bigsizeSize,2)
outLines = maindata.data(bigsizeSize,3)
outPosLines = maindata.data(bigsizeSize,4)

% extract the data we want
inData = inOutData(1:inLines,:);
inPosData = inData(1:inPosLines,:);
outData = inOutData(inLines+1:numLines,:);
outPosData = outData(1:outPosLines,:);

inScores = inPosData(:,4);
outScores = outPosData(:,4);
inScale = max(inScores) + 1;
outScale = max(outScores) + 1;
shiftstart = 10 + shift
shiftend = maxSeqLength + shift
BSBETASINGLECALC
	my $bsbetasingleUpper = <<BSBETASINGLEUPPER;
% upper box formatting
subplot('position', [0.08 0.7 0.84 0.15])
t = 1:1:maxSeqLength;
plot(t,0)
axis([1 maxSeqLength 0 inScale])
for kin = 1:inPosLines,
	inPosStart = inPosData(kin,1);
    inPosNormStart = (inPosStart - 1) / (maxSeqLength - 1);
	inPosEnd = inPosStart + inPosData(kin,2);
    inPosNormEnd = (inPosEnd - 1) / (maxSeqLength - 1);
    inPosX = [inPosNormStart inPosNormEnd];
	inPosScore = inPosData(kin,4);
    inPosNormScore = (inPosScore) / (inScale);
    inPosG = 1 - (inPosScore / inScale);
    inPosColor = [1 inPosG 0];
    inPosY = [inPosNormScore inPosNormScore];
	tline = (inPosStart:0.05:inPosEnd);
	line(tline,inPosScore,'LineWidth',2, 'Color', inPosColor);
    % annotation(gcf, 'arrow', inPosX, inPosY, 'Color', inPosColor);
end
BSBETASINGLEUPPER
	my $bsbetasingleLabel = <<BSBETASINGLELABEL;
set(gca, 'XTick',10:10:maxSeqLength)
set(gca, 'YTick',1:1:inScale)
set(gca, 'XTickLabel', shiftstart:10:shiftend)
xlabel('N-terminal residue')
ylabel('log score')
title('predicted single strands, odd residues buried')
BSBETASINGLELABEL
	my $bsbetasingleLower = <<BSBETASINGLELOWER;
% lower box formatting
subplot ('position',[0.08 0.35 0.84 0.15])
t = 1:0.05:maxSeqLength;
plot(t,0)
axis([1 maxSeqLength 0 inScale])
for kin = 1:outPosLines,
	outPosStart = outPosData(kin,1);
	outPosEnd = outPosStart + outPosData(kin,2);
	outPosScore = outPosData(kin,4);
    outPosG = 1 - (outPosScore / outScale);
    outPosColor = [1 outPosG 0];
	tline = (outPosStart:0.05:outPosEnd);
	line(tline,outPosScore,'LineWidth',2, 'Color', outPosColor);
end
BSBETASINGLELOWER
	my $bsbetapairBody = <<BSBETAPAIRBODY;
% now extract data
sourcefile = filename;
passedData = maindata(1,:)
maxColumns = passedData(1);
maxRows = passedData(2);
bigsizeVec = size(maindata);
bigsizeSize = bigsizeVec(1)
inOutData = maindata(2:bigsizeSize-1,:);
sizeVec = size(inOutData);
numLines = sizeVec(1)
inLines = maindata(bigsizeSize,1)
inPosLines = maindata(bigsizeSize,2)
outLines = maindata(bigsizeSize,3)
outPosLines = maindata(bigsizeSize,4)
antiParallel = maindata(bigsizeSize, 5)

% extract the data we want
inData = inOutData(1:inLines,:);
inPosData = inData(inLines-inPosLines+1:inLines,:);
outData = inOutData(inLines+1:numLines,:);
outPosData = outData(outLines-outPosLines+1:outLines,:);

inScores = inPosData(:,7);
outScores = outPosData(:,7);
inScale = max(inScores);
outScale = max(outScores);
twoScales = [inScale outScale];
fullScale = max(twoScales);
shiftstart = 10 + shift
shiftColumns = shift + maxColumns
shiftRows = shift + maxRows

% box formatting
t = 1:1:maxColumns;
subplot('position',[0.11 0.11 0.8 0.8])
plot(t,0)
axis([1 maxColumns 1 maxRows])
for kin = 1:inPosLines,
	inPosStartX = inPosData(kin,1);
    inPosNormStartX = (inPosStartX - 1) / (maxColumns - 1);
	inPosEndX = inPosStartX + inPosData(kin,2);
    inPosNormEndX = (inPosEndX - 1) / (maxColumns - 1);
    inPosX = [inPosNormStartX inPosNormEndX];
	inPosStartY = inPosData(kin,4) + (antiParallel * inPosData(kin, 5));
	inPosNormStartY = (inPosStartY - 1) / (maxRows - 1);
	inPosEndY = inPosStartY + ((1 - (2*antiParallel)) * inPosData(kin,5));
    inPosNormEndY = (inPosEndY - 1) / (maxRows - 1);
	inPosScore = inPosData(kin,7);
    inPosNormScore = (inPosScore) / (fullScale);
    inPosG = 1 - (inPosScore / fullScale);
    inPosColor = [1 inPosG 0];
    inPosY = [inPosStartY inPosEndY];
	tline = (inPosStartX:0.05:inPosEndX);
	tSlope = (inPosEndY - inPosStartY) / (inPosEndX - inPosStartX);
	tUnit = 0.05 * tSlope;
	tY = (inPosStartY:tUnit:inPosEndY);
	if (inPosData(kin,3) == 0) 
		line(tline,tY,'LineWidth',1, 'Color', inPosColor);
	else
		line(tline,tY,'LineWidth',1, 'Color', inPosColor,'LineStyle','--');
	end
    % annotation(gcf, 'arrow', inPosX, inPosY, 'Color', inPosColor);
end
BSBETAPAIRBODY
	my $bsbetapairLabel = <<BSBETAPAIRLABEL;
set(gca, 'XTick',10:10:maxColumns)
set(gca, 'YTick',10:10:maxRows)
set(gca, 'XTickLabel', shiftstart:10:shiftColumns)
set(gca, 'YTickLabel', shiftstart:10:shiftRows)

xlabel('residue number, upper strand')
ylabel('residue number, lower strand')
title('predicted pairwise strands')
BSBETAPAIRLABEL
	my $bsbetaTextChunk = <<BSBETATEXTCHUNK;

% tag image with file name and timestamp
textbox1 = annotation('textbox',[0 0.95 0.2 0.05])
set(textbox1,'String',filename)
set(textbox1,'FitHeightToText','on')
set(textbox1,'EdgeColor','none')

textbox2 = annotation('textbox',[0.8 0.95 0.2 0.05])
set(textbox2, 'String', datestr(now))
set(textbox2, 'FitHeightToText','on')
set(textbox2, 'EdgeColor','none')
set(textbox2, 'HorizontalAlignment','right')

textbox3 = annotation('textbox',[0 0.92 0.2 0.05])
set(textbox3,'String',optString)
set(textbox3,'FitHeightToText','on')
set(textbox3,'EdgeColor','none')
BSBETATEXTCHUNK
	my $renderPrepChunk = <<RENDERPREPCHUNK;

% make the renderer work right
set(gcf,'Renderer','Zbuffer');

% save file as $imgTag
output = [filename '.$imgTag']

RENDERPREPCHUNK
	#### generate rendering command using -i frobs ####
	my $renderChunk;
	my $renderJpgChunk = <<RENDERJPGCHUNK;
% saveas(gcf, output, 'jpg')
% exportfig(gcf, output, 'format', 'jpg', 'color', 'rgb')
% hgexport(gcf, output)
RENDERJPGCHUNK
	my $renderEpsChunk = <<RENDEREPSCHUNK;
% saveas(gcf, output, 'eps')
% exportfig(gcf, output, 'color', 'rgb')
% exportfig(gcf, output, 'width',4.5,'height',3,'Color','cymk') 
RENDEREPSCHUNK
	my $renderTiffChunk = <<RENDERTIFFCHUNK;
% saveas(gcf, output, 'tiff')
% print (gcf, '-dtiff', output)
RENDERTIFFCHUNK
	my $renderPdfChunk = <<RENDERPDFCHUNK;
saveas(gcf, output, 'pdf')
RENDERPDFCHUNK

	if ($imgTag eq "jpg") {		$renderChunk = $renderJpgChunk; }
	elsif ($imgTag eq "eps") {	$renderChunk = $renderEpsChunk; }
	elsif ($imgTag eq "tiff") { $renderChunk = $renderTiffChunk;}
	elsif ($imgTag eq "pdf") {	$renderChunk = $renderPdfChunk; }
	elsif ($imgTag eq "all") {
		# I'm actually using 'all' now for 'auto' and may actually switch it
								$renderChunk = $renderPdfChunk; }
	
	my $fontsizeLine = "set(gca, 'FontSize', 16)\n";
	
	### WHEW! Finally done with definitions, let's make some files
	my $bsplotsingleName = "bsplotsingle400$idtime.m";
	my $bsplotpairName =   "bsplotpair400$idtime.m";
	my $bsbetasingleName = "bsbetasingle400$idtime.m";
	my $bsbetapairName =   "bsbetapair400$idtime.m";
	
	open (PLOTSINGLE, ">$bsplotsingleName") or die "Cannot write the bsplotsingle file";
	open (PLOTPAIR,   ">$bsplotpairName"  ) or die "Cannot write the bsplotpair file  ";
	open (BETASINGLE, ">$bsbetasingleName") or die "Cannot write the bsbetasingle file";
	open (BETAPAIR,   ">$bsbetapairName"  ) or die "Cannot write the bsbetapair file  ";
	
	# generate bsplotsinglefile
	print PLOTSINGLE $bsplotsingleHeader;
	print PLOTSINGLE $dataSnarfChunk;
	print PLOTSINGLE $diaryChunk;
	print PLOTSINGLE $bsplotsingleCalc;
	print PLOTSINGLE $bsplotsingleUpperBox;
	print PLOTSINGLE $bsplotsingleHello;
	if ($largePrint > 0) {
		print PLOTSINGLE $fontsizeLine;
	}
	print PLOTSINGLE $bsplotsingleUpperBoxLabel;
	print PLOTSINGLE $bsplotsingleLowerBox;
	print PLOTSINGLE $bsplotsingleHello;
	if ($largePrint > 0) {
		print PLOTSINGLE $fontsizeLine;
	}
	print PLOTSINGLE $bsplotsingleLowerBoxLabel;
	print PLOTSINGLE $bsplotsingleText;
	print PLOTSINGLE $renderPrepChunk;
	print PLOTSINGLE $renderChunk;
	print PLOTSINGLE "diary off \n";
	close PLOTSINGLE;
	
	# generate bsplotpairfile
	print PLOTPAIR $bsplotpairHeader;
	print PLOTPAIR $dataSnarfChunk;
	print PLOTPAIR $diaryChunk;
	print PLOTPAIR $bsplotpairCalc;
	print PLOTPAIR $bsplotpairBox;
	if ($largePrint > 0) {
		print PLOTPAIR $fontsizeLine;
	}
	print PLOTPAIR $bsplotpairLabel;
	print PLOTPAIR $bsplotpairText;
	print PLOTPAIR $renderPrepChunk;
	print PLOTPAIR $renderChunk;
	print PLOTPAIR "diary off\n";
	close PLOTPAIR;
	
	# generate bsbetasinglefile
	print BETASINGLE $bsbetasingleHeader;
	print BETASINGLE $dataSnarfChunk;
	print BETASINGLE $diaryChunk;
	print BETASINGLE $bsbetasingleCalc;
	print BETASINGLE $bsbetasingleUpper;
	if ($largePrint > 0) {
		print BETASINGLE $fontsizeLine;
	}
	print BETASINGLE $bsbetasingleLabel;
	print BETASINGLE $bsbetasingleLower;
	if ($largePrint > 0) {
		print BETASINGLE $fontsizeLine;
	}
	print BETASINGLE $bsbetasingleLabel;
	print BETASINGLE $bsbetaTextChunk;
	print BETASINGLE $renderPrepChunk;
	print BETASINGLE $renderChunk;
	print BETASINGLE "diary off\n";
	close BETASINGLE;
	
	# generate bsbetapairfile
	print BETAPAIR $bsbetapairHeader;
	print BETAPAIR $dataSnarfChunk;
	print BETAPAIR $diaryChunk;
	print BETAPAIR $bsbetapairBody;
		if ($largePrint > 0) {
		print BETAPAIR $fontsizeLine;
	}
	print BETAPAIR $bsbetapairLabel;
	print BETAPAIR $bsbetaTextChunk;
	print BETAPAIR $renderPrepChunk;
	print BETAPAIR $renderChunk;
	print BETAPAIR "diary off\n";
	close BETAPAIR;
	
	system("sync");
	return $idTime;

	### Old rendering elements ###
	my $bsplotsingleRender = <<BSPLOTSINGLERENDER;
% make the renderer work right
% set(gcf,'Renderer','Zbuffer');

% save file as jpg and eps
outputtiff = [filename '.tiff'];
outputpdf = [filename '.pdf'];
outputjpg = [filename '.jpg'];
outputeps = [filename '.eps'];

set(gcf,'Renderer','Zbuffer');

% if (strcmp(imgTag,'eps'))
%	saveas(gcf, outputeps, 'eps')
% elseif (strcmp(imgTag,'jpg'))
%	saveas(gcf, outputjpg, 'jpg')
% elseif (strcmp(imgTag,'tiff'))
%	saveas(gcf, outputtiff, 'tiff')
% elseif (strcmp(imgTag,'pdf'))
% 	print(gcf, '-dpdf', outputpdf)
% elseif (strcmp(imgTag,'all'))
%	print(gcf, '-dpdf', outputpdf)
% saveas(gcf, outputeps, 'eps')
%	saveas(gcf, outputjpg, 'jpg')
%	saveas(gcf, outputtiff, 'tiff')
% endif

% exportfig (gcf, outputeps, 'color', 'rgb')
% exportfig (gcf, outputeps, 'format', 'jpg', 'color', 'rgb') 
print (gcf, '-dpdf', outputpdf)
% exportfig(gcf, outputeps,'width',4.5,'height',3,'Color','cmyk')
% print (gcf, '-dtiff', outputtiff)
% print (gcf, '-djpeg', outputjpg)
% hgexport (gcf, outputjpg)

diary off
BSPLOTSINGLERENDER
	my $bsplotpairRender = <<BSPLOTPAIRRENDER;
% save file as jpg and eps
outputtiff = [filename '.tiff'];
outputpdf = [filename '.pdf'];
outputjpg = [filename '.jpg'];

set(gcf,'Renderer','Zbuffer');

% saveas(gcf, outputjpg, 'jpg')
% saveas(gcf, outputeps, 'eps')
% exportfig (gcf, outputeps, 'color', 'rgb')
% exportfig (gcf, outputeps, 'format', 'jpg', 'color', 'rgb') 
print (gcf, '-dpdf', outputpdf)
print (gcf, '-dtiff', outputtiff)
print (gcf, '-djpeg', outputjpg)
%hgexport (gcf, outputjpg)

filename

diary off
BSPLOTPAIRRENDER
	my $bsbetasingleRender = <<BSBETASINGLERENDER;
% make the renderer work right
% set(gcf,'Renderer','Zbuffer');

% save file as jpg and eps
outputtiff = [filename '.tiff'];
outputpdf = [filename '.pdf'];
outputjpg = [filename '.jpg'];
outputeps = [filename '.eps'];

set(gcf,'Renderer','Zbuffer');

% saveas(gcf, outputeps, 'eps')
% saveas(gcf, outputjpg, 'jpg')
% saveas(gcf, outputtiff, 'tiff')
% saveas(gcf, outputpdf, 'pdf')
% exportfig (gcf, outputeps, 'color', 'rgb')
% exportfig (gcf, outputeps, 'format', 'jpg', 'color', 'rgb') 
print (gcf, '-dpdf', outputpdf)
% print (gcf, '-dtiff', outputtiff)
% print (gcf, '-djpeg', outputjpg)
% hgexport (gcf, outputjpg)

diary off
BSBETASINGLERENDER
	my $bsbetapairRender = <<BSBETAPAIRRENDER;
% make the renderer work right
% set(gcf,'Renderer','Zbuffer');

% save file as jpg and eps
outputtiff = [filename '.tiff'];
outputpdf = [filename '.pdf'];
outputjpg = [filename '.jpg'];
outputeps = [filename '.eps'];

set(gcf,'Renderer','Zbuffer');

% saveas(gcf, outputeps, 'eps')
% saveas(gcf, outputjpg, 'jpg')
% saveas(gcf, outputtiff, 'tiff')
% saveas(gcf, outputpdf, 'pdf')
% exportfig (gcf, outputeps, 'color', 'rgb')
% exportfig (gcf, outputeps, 'format', 'jpg', 'color', 'rgb') 
print (gcf, '-dpdf', outputpdf)
% print (gcf, '-dtiff', outputtiff)
% print (gcf, '-djpeg', outputjpg)
% hgexport (gcf, outputjpg)

diary off
BSBETAPAIRRENDER
}

##################################################################
### Z. Plain Old Documentation                                 ###
##################################################################

=pod
=head1 NAME

 Betascan: a program to predict beta-sheets from sequence alone using pairwise probabilistic modeling

=head1 VERSION HISTORY

 2008-10-27 final cleanup for web download release
 2008-10-21 add batch fallthrough: failure to find strands now does not crash program
 2008-03-14 update to documentation
 2007-10-18 add zip option; move correction into initial output
 2007-09-14 betascan web output 
 2007-08-04 integrate matlab files into perl code and implement image switch
 2006-09-14	reimplement pairwise detector with improved dynamic programming
 2006-08-24 add two-sequence capability
 2006-08-21 add additional weighting for tyrosine, glutamine, asparagine pairings
 2006-08-09 calculations with length disabled (may be useful for stitcher)
 2006-07-30 antiparallel calculations
 2006-07-30 even more switches and access points added; considering changing defaults on many switches
 2006-07-27 rearranged architecture; many more switches and access points added 
 2006-07-26 transferred calcProbs out of code4.cpp and into perl script
 2006-07-24 started work on corrector
 2006-07-17 extensive commenting and organization
 2006-07-14 graphing capability added for maxima scores down to -1
 2006-07-13 maxima calculators and graphers added
 2006-07-11 log-odds score calculations added; color settings adjusted
 2006-07-10 normalization by length
 2006-07-07 normalization by orientation 
 2006-07-05 bug found: was calculating beta strands that hung off the end of the chart.
 2006-07-03 -m and -M switches: minimum and maximum loops between strands now frobbable on command line
 2006-07-01 bsplotpair.m debugged and incorporated. now generating on-the-fly pairwise graphs
 2006-06-29 bsplotsingle.m and code4.cpp now use data from switches
 2006-06-28 switches implemented. input file name, beta strand length, and length of file to be graphed
 2006-06-27 seperate perl files now incorporated as subroutines
 2006-06-26 matlab invokation debugged. bsplotsingle.m now working
 2005-05-20 incorporating seperate drivers into a single master driver perl script
 2005-05-20 see archived individual perl scripts for previous version histories

=head1 DESCRIPTION

Betascan uses the pairwise beta-structure databases compiled for Betawrap (latest version: McDonnell et al., 2006)
to calculate the probability of

* (a) a beta strand of given length, position, and orientation forming (the singleton probability) and 

* (b) two beta strands of given length, positions, and orientation forming part of a beta-sheet.
We calculate these matters progressively over a specified length of sequence and normalize:

* (c) by length (according to a Poisson model fitted to the data of Penel et al., 2003)
and generate log-odds scores by calculating a null model where all conditional probabilities are 0.05.
Next, we identify all local maxima in singleton and pairwise scores over a space populated by the calculated scores
and with possible transitions defined by the physically possible changes in beta-strand (or beta-sheet) folding. 
Beta-strand and sheet scores for the all possibilities and/or for local maxima only may be graphed.

=head2 Dependencies

Betascan consists of the following files:

 betascan.pl
	Main program
 BSweights.txt
 BSprobs.txt
	Probability tables
	
Betascan requires the following installations:

 Perl 6 or later
 Math::Matlab::Local
 Getopt::Std
 MATLAB (7.0.1 or later)

=head2 Input files

 BETASCAN requires one input file: a FASTA format containing one or more sequences. 
 IMPORTANT: The name of the protein(s) MUST be non-identical in the first word!

=head2 Output files

BETASCAN generates the following data files:

 [Name of protein] prion single strand [in-out|out-in] [strand length] residue.txt
	singleton predictions for each orientation and strand length
 [Name of protein] prion single strand complete swap.txt
	arranged for singleton graphing
 [Name of protein] prion forward [in-out|out-in] [strand length] residue.txt
	pairwise predictions of sheets, with upper strand as prior
 [Name of protein] 3dranged prion forward [in-out|out-in] [strand length] residue.txt
	arranged for pairwise graphing
 [Name of protein] prion combined [in-out|out-in] [strand length] residue.txt
	pairwise predictions of sheets multiplied by singleton predictions of the upper sheet
 [Name of protein] 3dranged prion forward [in-out|out-in] [strand length] residue.txt
	arranged for pairwise graphing
 [Name of protein] prion single strand complete maxima.txt
	singleton prediction local maxima, arranged for graphing
 [Name of protein] prion combined scores complete maxima.txt
	pairwise prediction local maxima, arranged for graphing

 singleInFreqs.txt: singleton interior frequencies for the twenty amino acids
 singleOutFreqs.txt: singleton exterior frequencies for the twenty amino acids
 
 bsplotsingle400(run ID).m
 bsplotpair400(run ID).m
 bsbetasingle400(run ID).m
 bsbetapair400(run ID).m
	Matlab files custom-tailored to generate desired plots

 Error files, one for each protein and one for each pairwise maxima calculation.

 BETASCAN also generates an image file for each arranged file, in pdf format.
 BETASCAN may also generate in eps, tiff, or jpg formats, as controlled by the -i switch, but compatibility is not guaranteed!
 Each invocation of MATLAB generates a diary for debugging.

=head2 Switches

The following switches work with BETASCAN:

=head3 Mandatory inputs:

 -n B<inputfilename:> takes the input FASTA file.
 -b B<betalength:> the maximum length of a beta strand in the model. 
	    Must be integer greater than 2. 
 -l B<seqlength:> the length of the sequence. Only the first I<seqlength> residues will be graphed.

=head3 Optional inputs:

 -m B<minloop:> the minimum distance between two possibly paired beta strands. 
        Current default is 0.
 -M B<maxloop:> the maximum distance between two possibly paired beta strands. 
        Current default is 35. 
	    Setting to 0 will force checking of all possible pairings (i.e., all interchain contacts).
 -z B<startPoint:> the number of the first residue in the sequence, if not 1.
 -c B<modelname:> the amino acid composition model to be used. 
        Must be I<none:>, I<null:>, or one of the models in BSweights.txt (or specified by -w). 
	 	Defaults to I<allproteins:>.
 -i B<format:> requested output format for images. 
        Currently available: pdf, tiff, jpg, all.

=head3 Location inputs:

 -j B<jobname:> the name of the job, to be used in assigning file and directory names.
 -G B<matlabLoc:> the location of MATLAB on the hard drive, if not in the default location.
		Default location is "/Applications/MATLAB701/bin/matlab".
 -C B<cacheAddr:> the cache address, to be used to place BETASCAN output in a temporary directory.

=head3 Alternate prior data files:

 -w B<weightfilename:> takes the file with amino acid composition models. 
        Defaults to BSweights.txt.
 -p B<probfilename:> takes the file with conditional probability tables. 
        Defaults to BSprobs.txt.

=head3 Calculation switches:

 -L: turn off normalization by length
 -e: turn on maxima filter to remove less likely maxima
 -y: use bonuses for tyrosine, asparagine, and glutaimne side-chain stacking interactions.

=head3 Output switches:

 -1: only generate singleton graphs.
 -2: only generate pairwise graphs.
 -g: turn off singleton and pairwise image generation.
 -s: skip all pre-maxima calculations and graphs.
 -F: generate only most-used graphs.
 -f: turn off finding and graphing local maxima.
 -d: do not move data files and images to their own subdirectories at the end of calculations.
 -h: turn off HTML file with links to data files.
 -T: enlarge all graph text, suitable for generating images for publication.
 -Z: generate zipped archive of all outputs

=head3 Debug switches:

 -D: explicit debugging. I<Warning::> this will cause the debug file to become VERY large (>25 MB).
 -t: include timestamps in debugging messages.

=head3 Deprecated experimental switches:

 -a B<seqID:> the secondary sequence to run all sequences against.
		Must exist within the input FASTA file.
		Used only to do hetero pairwise runs (runs of one sequence forming beta-structure with another).
 -o: turn on normalization by orientation.
 -A: calculate antiparallel pairwise probabilities.
 -N B<normParam:> change parameter of length normalization.
 

=head1 AUTHOR

 Allen W. Bryan, Jr., lead author and implementer
 Bonnie Berger, Susan Lindquist, advisors
 Matt Menke, Andrew McDonnell, coders and implementers
 Lenore Cowen, massive amounts of good advice and paper-writing help
 Additional immense amounts of advice and test data from the Susan Lindquist lab and Thronson Luehrs lab
 Random additional support by the Bonnie Berger lab
 Help with Matlab debugging from Shannon Weiland and Sepideh Baghaii
 Parts of the amino acid propensity tables provided by Oliver King
 BETASCAN conditional probability tables updated by Nathan Palmer
 Website maintenance and debugging by Kenny Lu
 
 A production of:
 * Massachusetts Institute of Technology
	* Computer Science and Artificial Intelligence Laboratory
	* Department of Mathematics (Applied)
	* Department of Biology
 * Harvard/MIT Divison of Health Sciences and Technology
 * Whitehead Institute for Biomedical Research
 * Tufts University
	* Computer Science Department
 
 With generous support by:
 * National Institutes of Health
	* Informatics for Integrating Biology & the Bedside, grant # U54-LM008748
	* grant # R01-GM25974
 * National Science Foundation
    * award # 0428715
 
 For relevant legal verbage, talk to their lawyers...
 
 ALL INQUIRIES ARE DIRECTED TO: betascan@csail.mit.edu
 
# #include <disclaimer.h>

=cut
